import copy
import re
from typing import Any, ClassVar, cast

import pytest
from eth_pydantic_types import HexBytes, HexBytes32
from eth_typing import HexAddress, HexStr
from ethpm_types import ContractType, ErrorABI
from ethpm_types.abi import ABIType, EventABI, MethodABI
from evm_trace import CallTreeNode, CallType

from ape.api.networks import ForkedNetworkAPI, NetworkAPI
from ape.exceptions import CustomError, DecodingError, NetworkError, NetworkNotFoundError
from ape.types.address import AddressType
from ape.types.gas import AutoGasLimit
from ape.types.units import CurrencyValueComparable
from ape.utils.misc import DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT, LOCAL_NETWORK_NAME
from ape_ethereum.ecosystem import BLUEPRINT_HEADER, BaseEthereumConfig, Block, Ethereum
from ape_ethereum.trace import TransactionTrace
from ape_ethereum.transactions import (
    DynamicFeeTransaction,
    Receipt,
    SetCodeTransaction,
    SharedBlobReceipt,
    SharedBlobTransaction,
    StaticFeeTransaction,
    TransactionType,
)

LOG = {
    "removed": False,
    "logIndex": "0x0",
    "transactionIndex": "0x0",
    "transactionHash": "0x74dd040dfa06f0af9af8ca95d7aae409978400151c746f55ecce19e7356cfc5a",
    "blockHash": "0x2c99950b07accf3e442512a3352a11e6fed37b2331de5f71b7743b357d96e4e8",
    "blockNumber": "0xa946ac",
    "address": "0x274b028b03a250ca03644e6c578d81f019ee1323",
    "data": "0xabffd4675206dab5d04a6b0d62c975049665d1f512f29f303908abdd20bc08a100000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000060000000000000000000000000000000000000000000000000000000000000000744796e616d696300000000000000000000000000000000000000000000000000",
    "topics": [
        "0xa84473122c11e32cd505595f246a28418b8ecd6cf819f4e3915363fad1b8f968",
        "0x0000000000000000000000000000000000000000000000000000000000000006",
        "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d",
    ],
}
CUSTOM_ECOSYSTEM_NAME = "custom-ecosystem"


def make_method_abi(name: str, inputs=None, outputs=None) -> MethodABI:
    inputs = inputs or []
    outputs = outputs or []
    return MethodABI.model_validate(
        {
            "type": "function",
            "name": name,
            "stateMutability": "nonpayable",
            "inputs": inputs,
            "outputs": outputs,
        }
    )


@pytest.fixture
def event_abi(vyper_contract_instance):
    return vyper_contract_instance.NumberChange.abi


@pytest.fixture
def configured_custom_ecosystem(custom_networks_config_dict, project, custom_network_name_0):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["ecosystem"] = CUSTOM_ECOSYSTEM_NAME

    # Also ensure we can configure the custom ecosystem as if
    # it were from a plugin.
    data[CUSTOM_ECOSYSTEM_NAME] = {"default_network": custom_network_name_0}

    with project.temp_config(**data):
        yield


def test_name(ethereum):
    assert ethereum.name == "ethereum"


def test_request_header(ethereum):
    actual = ethereum.request_header
    expected = {"User-Agent": "ape-ethereum"}
    assert actual == expected


def test_request_header_subclass():
    class L2(Ethereum):
        name: str = "l2"

    l2 = L2()
    actual = l2.request_header
    expected = {"User-Agent": "ape-l2"}
    assert actual == expected


def test_name_when_custom(configured_custom_ecosystem, networks):
    ecosystem = networks.get_ecosystem(CUSTOM_ECOSYSTEM_NAME)
    actual = ecosystem.name
    expected = CUSTOM_ECOSYSTEM_NAME
    assert actual == expected


@pytest.mark.parametrize(
    "address",
    (
        "0x63953eB1B3D8DB28334E7C1C69456C851F934199".lower(),
        0x63953EB1B3D8DB28334E7C1C69456C851F934199,
    ),
)
def test_decode_address(ethereum, address):
    expected = "0x63953eB1B3D8DB28334E7C1C69456C851F934199"
    actual = ethereum.decode_address(address)
    assert actual == expected


def test_encode_address(ethereum):
    raw_address = "0x63953eB1B3D8DB28334E7C1C69456C851F934199"
    address = AddressType(HexAddress(HexStr(raw_address)))
    actual = ethereum.encode_address(address)
    assert actual == raw_address


def test_encode_calldata(ethereum, address):
    abi = make_method_abi(
        "callMe",
        inputs=[
            ABIType(name="a", type="bytes4"),
            ABIType(name="b", type="address"),
            ABIType(name="c", type="uint256"),
            ABIType(name="d", type="bytes4[]"),
        ],
    )
    byte_array = ["0x456", "0x678"]
    values = ("0x0123", address, HexBytes(55), byte_array)

    actual = ethereum.encode_calldata(abi, *values)
    expected = HexBytes(
        # 0x123
        "0123000000000000000000000000000000000000000000000000000000000000"
        # address
        "000000000000000000000000d8da6bf26964af9d7eed9e03e53415d37aa96045"
        # HexBytes(55)
        "0000000000000000000000000000000000000000000000000000000000000037"
        # byte_array
        "0000000000000000000000000000000000000000000000000000000000000080"
        # length
        "0000000000000000000000000000000000000000000000000000000000000002"
        # first bytes4
        "0456000000000000000000000000000000000000000000000000000000000000"
        # second bytes4
        "0678000000000000000000000000000000000000000000000000000000000000"
    )
    assert actual == expected


@pytest.mark.parametrize(
    "sequence_type,item_type",
    [
        (list, str),
        (tuple, str),
        (list, HexBytes),
        (tuple, HexBytes),
    ],
)
def test_encode_calldata_byte_array(ethereum, sequence_type, item_type):
    """
    Tests against a bug where we could not pass a tuple of HexStr
    for a byte-array.
    """
    abi = make_method_abi(
        "mint", inputs=[{"name": "leaf", "type": "bytes32"}, {"name": "proof", "type": "bytes32[]"}]
    )
    hexstr_array = sequence_type(
        (
            item_type("0xfadbd3"),
            item_type("0x9c6b2c"),
            item_type("0xc5d246"),
        )
    )
    actual = ethereum.encode_calldata(abi, "0x0123", hexstr_array)
    assert isinstance(actual, bytes)


def test_encode_calldata_nested_structs(ethereum):
    abi = make_method_abi(
        "check",
        inputs=[
            ABIType(
                name="data",
                type="tuple",
                components=[
                    ABIType(
                        name="tokenIn", type="address", components=None, internal_type="address"
                    ),
                    ABIType(
                        name="amountIn", type="uint256", components=None, internal_type="uint256"
                    ),
                    ABIType(
                        name="minAmountOut",
                        type="uint256",
                        components=None,
                        internal_type="uint256",
                    ),
                    ABIType(
                        name="path",
                        type="tuple",
                        components=[
                            ABIType(
                                name="pairBinSteps",
                                type="uint256[]",
                                components=None,
                                internal_type="uint256[]",
                            ),
                            ABIType(
                                name="versions",
                                type="uint8[]",
                                components=None,
                                internal_type="enum SwapWrapper.Version[]",
                            ),
                            ABIType(
                                name="tokenPath",
                                type="address[]",
                                components=None,
                                internal_type="address[]",
                            ),
                        ],
                        internal_type="struct SwapWrapper.Path",
                    ),
                ],
                internal_type="struct SwapWrapper.SwapData",
            )
        ],
        outputs=[ABIType(name="", type="bool", components=None, internal_type="bool")],
    )
    calldata = {
        "tokenIn": "0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7",
        "amountIn": 1000000000000000000,
        "minAmountOut": 10000000,
        "path": {
            "pairBinSteps": [0],
            "versions": [0],
            "tokenPath": [
                "0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7",
                "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E",
            ],
        },
    }
    actual = ethereum.encode_calldata(abi, calldata)
    expected = HexBytes(
        "0x000000000000000000000000000000000000000000000000000000000000002"
        "0000000000000000000000000b31f66aa3c1e785363f0875a1b74e27b85fd66c7"
        "0000000000000000000000000000000000000000000000000de0b6b3a76400000"
        "00000000000000000000000000000000000000000000000000000000098968000"
        "00000000000000000000000000000000000000000000000000000000000080000"
        "00000000000000000000000000000000000000000000000000000000000600000"
        "0000000000000000000000000000000000000000000000000000000000a000000"
        "000000000000000000000000000000000000000000000000000000000e0000000"
        "00000000000000000000000000000000000000000000000000000000010000000"
        "00000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000001000000000"
        "00000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000200000000000"
        "0000000000000b31f66aa3c1e785363f0875a1b74e27b85fd66c7000000000000"
        "000000000000b97ef9ef8734c71904d8002f8b6bc66dd9c48a6e"
    )
    assert actual == expected


def test_block_handles_snake_case_parent_hash(eth_tester_provider, sender, receiver):
    # Transaction to change parent hash of next block
    sender.transfer(receiver, "1 gwei")

    # Replace 'parentHash' key with 'parent_hash'
    latest_block = eth_tester_provider.get_block("latest")
    latest_block_dict = eth_tester_provider.get_block("latest").model_dump()
    latest_block_dict["parent_hash"] = latest_block_dict.pop("parentHash")

    redefined_block = Block.model_validate(latest_block_dict)
    assert redefined_block.parent_hash == latest_block.parent_hash


def test_transaction_acceptance_timeout(project, networks):
    assert (
        networks.provider.network.transaction_acceptance_timeout
        == DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT
    )
    new_value = DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT + 1
    timeout_config = {"ethereum": {"local": {"transaction_acceptance_timeout": new_value}}}
    with project.temp_config(**timeout_config):
        assert networks.provider.network.transaction_acceptance_timeout == new_value


@pytest.mark.parametrize("removed", [True, False])
def test_decode_logs(ethereum, vyper_contract_instance, removed):
    abi = vyper_contract_instance.NumberChange.abi
    log = {**LOG, "removed": removed}
    result = [x for x in ethereum.decode_logs([log], abi)]
    assert len(result) == 1
    model_dump = result[0].model_dump()
    assert model_dump == {
        "event_name": "NumberChange",
        "contract_address": "0x274b028b03A250cA03644E6c578D81f019eE1323",
        "event_arguments": {
            "newNum": 6,
            "dynIndexed": HexBytes(
                "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d"
            ),
            "b": HexBytes("0xabffd4675206dab5d04a6b0d62c975049665d1f512f29f303908abdd20bc08a1"),
            "prevNum": 0,
            "dynData": "Dynamic",
        },
        "transaction_hash": "0x74dd040dfa06f0af9af8ca95d7aae409978400151c746f55ecce19e7356cfc5a",
        "block_number": 11093676,
        "block_hash": "0x2c99950b07accf3e442512a3352a11e6fed37b2331de5f71b7743b357d96e4e8",
        "log_index": 0,
        "transaction_index": 0,
        "removed": removed,
    }


def test_decode_logs_empty_list(ethereum, event_abi):
    actual = [x for x in ethereum.decode_logs([], event_abi)]
    assert actual == []


def test_decode_logs_with_struct_from_interface(ethereum):
    abi = EventABI.model_validate(
        {
            "type": "event",
            "name": "ConditionalOrderCreated",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address", "indexed": True},
                {
                    "name": "params",
                    "type": "tuple",
                    "components": [
                        {
                            "name": "handler",
                            "type": "address",
                            "internalType": "contract IConditionalOrder",
                        },
                        {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                        {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                    ],
                    "internalType": "struct IConditionalOrder.ConditionalOrderParams",
                    "indexed": False,
                },
            ],
            "anonymous": False,
        }
    )
    logs = [
        {
            "address": "0xfdafc9d1902f4e0b84f65f49f244b32b31013b74",
            "blockHash": "0xec1b8e18412dd91114c0b3ea03c41d02243d60cd05661861df5765fb5ec462c6",
            "blockNumber": "0x11101cc",
            "data": "0x00000000000000000000000000000000000000000000000000000000000000200000000000000000000000006cf1e9ca41f7611def408122793c358a3d11e5a5000000000000000000000000000000000000000000000000000000189e491e96000000000000000000000000000000000000000000000000000000000000006000000000000000000000000000000000000000000000000000000000000001400000000000000000000000006b175474e89094c44da98b954eedeac495271d0f000000000000000000000000c02aaa39b223fe8d0a0e5c4f27ead9083c756cc2000000000000000000000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd00000000000000000000000000000000000000000000018eb03406b6af33c8c3000000000000000000000000000000000000000000000000345d7669a1e8815700000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000012c00000000000000000000000000000000000000000000000000000000000000007b9bc44550e2cc369010139a1d3a5a6ec6d65910379e0ee35a7c3cdf24cb12a7",
            "logIndex": "0x97",
            "removed": False,
            "topics": [
                "0x2cceac5555b0ca45a3744ced542f54b56ad2eb45e521962372eef212a2cbf361",
                "0x000000000000000000000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd",
            ],
            "transactionHash": "0x55859e74e591100e0f37f146abc0dbebe2252eaf5f533319ca77427745b85ceb",
            "transactionIndex": "0x45",
        }
    ]
    actual = list(ethereum.decode_logs(logs, abi))
    assert len(actual) == 1
    assert actual[0].owner == "0xE7602Ca44f83a5E9Ba8BD14125dDcb295f3D63BD"
    assert actual[0].params == [
        "0x6cf1e9ca41f7611def408122793c358a3d11e5a5",
        HexBytes("0x000000000000000000000000000000000000000000000000000000189e491e96"),
        HexBytes(
            "0x0000000000000000000000006b175474e89094c44da98b954eedeac495271d0f0000000"
            "00000000000000000c02aaa39b223fe8d0a0e5c4f27ead9083c756cc20000000000000000"
            "00000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd0000000000000000000000000"
            "0000000000000000000018eb03406b6af33c8c30000000000000000000000000000000000"
            "00000000000000345d7669a1e881570000000000000000000000000000000000000000000"
            "0000000000000000000000000000000000000000000000000000000000000000000000000"
            "0000000000020000000000000000000000000000000000000000000000000000000000000"
            "12c00000000000000000000000000000000000000000000000000000000000000007b9bc4"
            "4550e2cc369010139a1d3a5a6ec6d65910379e0ee35a7c3cdf24cb12a7"
        ),
    ]


def test_decode_block_when_hash_is_none(ethereum):
    # When using some providers, such as hardhat, the hash of the pending block is None
    block_data_with_none_hash: dict[str, Any] = {
        "number": None,
        "hash": None,
        "parentHash": HexBytes(
            "0xcb94e150c06faee9ab2bf12a40b0937ac9eab1879c733ebe3249aafbba2f80b1"
        ),
        "nonce": None,
        "mixHash": None,
        "sha3Uncles": HexBytes(
            "0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347"
        ),
        "logsBloom": None,
        "transactionsRoot": HexBytes(
            "0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421"
        ),
        "stateRoot": HexBytes("0x8728474146565003152f9cee496de043fd68566dabdb06116a0d5bfc63e1a5a9"),
        "receiptsRoot": HexBytes(
            "0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421"
        ),
        "miner": "0xC014BA5EC014ba5ec014Ba5EC014ba5Ec014bA5E",
        "difficulty": 131072,
        "totalDifficulty": 131073,
        "extraData": HexBytes("0x"),
        "size": 513,
        "gasLimit": 0,
        "gasUsed": 0,
        "timestamp": 1660932629,
        "transactions": [],
        "uncles": [],
        "baseFeePerGas": 0,
    }
    actual = ethereum.decode_block(block_data_with_none_hash)
    assert actual.hash is None
    assert actual.gas_limit == 0


def test_decode_block_with_hex_values(ethereum):
    """
    When using WS providers directly, the values are all hex strings (more raw).
    This test ensures we can still decode blocks that are in this form.
    """
    raw_block = {
        "baseFeePerGas": "0x62d76fae2",
        "difficulty": "0x0",
        "extraData": "0x506f776572656420627920626c6f58726f757465",
        "gasLimit": "0x1c9c380",
        "gasUsed": "0x91b5e5",
        "hash": "0x925cac44d8bac5df3e3eeed6379b3924d6769a054b3c4079899c1d9b442a4041",
        "logsBloom": "0x7823d304ab8c72f91270a023c11482f341a08a00c91b00f8859906c8e9f22e04b19d311e422cfa79d8107e2b901fc7d9c62fcd2d8a673802f500f4a023693525a2ce66481cc51d1a4e97422b8424912b8c719c6821589a240d5015c5cbe4de46b3b60ad43621cd2f540119b3c938d88f011a49d755a31cc415f34896596912d40b0397d0c544571704ea001c56303f02086e16a3a3b7168f443d224034102ca4c6b21456932dae401c8da7c51d36ae3804c8a084811802c708ae94b7010302100700a28f156d2ca2d61420320e82e8b9927b626bc0ec003dfa09009e3d1668798fb0f8c81a205eb100ee8ea2e20c6041b2d8b2b01bd4d14e8acfaa317049d654",
        "miner": "0x6b3a8798e5fb9fc5603f3ab5ea2e8136694e55d0",
        "mixHash": "0x0493af8ac5b7de66c7660c22a703a54069cbd561a2767537eaef994bd77e084b",
        "nonce": "0x0000000000000000",
        "number": "0x11234df",
        "parentHash": "0x47651c13bda353d1bbc610bbdb8a9aa1436629f05f8a32a27f143200da698df8",
        "receiptsRoot": "0xa5abe7eeac74aa03ea82421fcef9b74d1822cfe0ee4b29a8dc01eef31e7d0019",
        "sha3Uncles": "0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347",
        "size": "0x1cf02",
        "stateRoot": "0x59eb1dfb098361d1f74dffc8436da680715f671dc3f02be28f4db3f44b1a09bf",
        "timestamp": "0x64e4b0cb",
        "transactionsRoot": "0xf2e3c3e4fbe06ff882f3deb8a7ec8cbbe09e55b2c995d93dd0a36d5e843f3efc",
    }
    actual = ethereum.decode_block(raw_block)
    assert actual.difficulty == 0


def test_decode_logs_topics_not_first(ethereum):
    """
    Tests against a condition where we could not decode logs if the
    topics were not defined first.
    """
    abi_dict = {
        "anonymous": False,
        "inputs": [
            {"indexed": False, "internalType": "address", "name": "sender", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "owner", "type": "address"},
            {"indexed": True, "internalType": "int24", "name": "tickLower", "type": "int24"},
            {"indexed": True, "internalType": "int24", "name": "tickUpper", "type": "int24"},
            {"indexed": False, "internalType": "uint128", "name": "amount", "type": "uint128"},
            {"indexed": False, "internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "amount1", "type": "uint256"},
        ],
        "name": "Mint",
        "type": "event",
    }
    abi = EventABI.model_validate(abi_dict)
    logs = [
        {
            "address": "0x3416cf6c708da44db2624d63ea0aaef7113527c6",
            "blockHash": "0x488f23ba55f64bf1aac02ee7278b70c3f4bb2fb57b8aaa6ab3b481f1809f18ea",
            "blockNumber": "0xcfa869",
            "data": "0x000000000000000000000000c36442b4a4522e871399cd717abdd847ab11fe8800000000000000000000000000000000000000000000000000821d6b102660fb000000000000000000000000000000000000000000000000000009fde8545338000000000000000000000000000000000000000000000000000003548cf35031",
            "logIndex": "0x7a",
            "removed": False,
            "topics": [
                "0x7a53080ba414158be7ec69b987b5fb7d07dee101fe85488f0853ae16239d0bde",
                "0x000000000000000000000000c36442b4a4522e871399cd717abdd847ab11fe88",
                "0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffc",
                "0x0000000000000000000000000000000000000000000000000000000000000004",
            ],
            "transactionHash": "0xf093a630478562d03e3b2476a5b0551609722747d442a462579fa02e1332a941",
            "transactionIndex": "0x41",
        }
    ]
    actual = list(ethereum.decode_logs(logs, abi))
    assert actual


def test_decode_receipt(eth_tester_provider, ethereum):
    receipt_data = {
        "provider": eth_tester_provider,
        "required_confirmations": 0,
        "blockHash": HexBytes("0x8988adc8a0b346526f1d769841b7464e38a282b25ed346f1f810be8cb7393bc2"),
        "blockNumber": 2,
        "from": "0x318b469BBa396AEc2C60342F9441be36A1945174",
        "gas": 1764431,
        "hash": HexBytes("0x58d086e6b7ecf55c92a5fe420d870ff2eebe99824a711a3cdbe29f497a0e534c"),
        "input": " ",
        "nonce": 1,
        "to": None,
        "transactionIndex": 0,
        "value": 0,
        "v": 1,
        "r": HexBytes("0xfa532141efbfa5bbeaa542cf68bfe5df97dd18b696b8a183f74714758a646238"),
        "s": HexBytes("0x1b80edf76bb6f0295ff847c1b99eb2928ea3533fe41125cfe0fa17fdc634b938"),
        "type": 2,
        "accessList": [],
        "chainId": 31337,
        "gasPrice": 0,
        "maxFeePerGas": 0,
        "maxPriorityFeePerGas": 0,
        "transactionHash": HexBytes(
            "0x58d086e6b7ecf55c92a5fe420d870ff2eebe99824a711a3cdbe29f497a0e534c"
        ),
        "cumulativeGasUsed": 1764431,
        "gasUsed": 1764431,
        "contractAddress": "0xC072c85922b7233998B2Ab990fEFdE80218Ca63F",
        "logs": [],
        "logsBloom": HexBytes(
            "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        ),
        "status": 1,
        "effectiveGasPrice": 0,
    }
    receipt = ethereum.decode_receipt(receipt_data)
    assert type(receipt) is Receipt  # NOTE: Purposely not using isinstance()

    # Tests against bug where input data would come back improperly
    assert receipt.data == HexBytes(
        " "
    )


def test_decode_receipt_from_etherscan(eth_tester_provider, ethereum):
    receipt = ethereum.decode_receipt(
        {
            "blockNumber": "11291970",
            "timeStamp": "1661846925",
            "hash": "0x5780b43d819035ed1fa079171bdce7f0bbeaa6b01f201f8985d279a66cfc6844",
            "nonce": "0",
            "blockHash": "0x3175f953c1da4bf3d15b853dae4a150ae44e2e71380936463e89142c12961968",
            "transactionIndex": "31",
            "from": "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
            "to": "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512",
            "value": "0",
            "gas": "4712388",
            "gasPrice": "1499999989",
            "isError": "0",
            "input": "0x608060405234801561000f575f80fd5b506101248061001d5f395ff3fe6080604052348015600e575f80fd5b50600436106026575f3560e01c8063f207564e14602a575b5f80fd5b6039603536600460b4565b603b565b005b6040516390b5561d60e01b81526004810182905273274b028b03a250ca03644e6c578d81f019ee1323906390b5561d90602401602060405180830381865af41580156088573d5f803e3d5ffd5b505050506040513d601f19601f8201168201806040525081019060aa919060ca565b60b1575f80fd5b50565b5f6020828403121560c3575f80fd5b5035919050565b5f6020828403121560d9575f80fd5b8151801515811460e7575f80fd5b939250505056fea2646970667358221220050eefcb4ae3417fbe131ccad8d577d9488919884d4c4a4d0bf8b9c46ce640f664736f6c63430008150033",
            "contractAddress": "",
            "cumulativeGasUsed": "8978461",
            "gasUsed": "257131",
            "methodId": "0xf926657e",
            "functionName": "setup(address _reputationOracle, address _recordingOracle, uint256 _reputationOracleStake, uint256 _recordingOracleStake, string _url, string _hash)",
            "required_confirmations": "13703",
            "status": "1",
            "chainId": 1,
        }
    )
    assert type(receipt) is Receipt  # NOTE: Purposely not using isinstance()
    assert receipt.type == 0
    assert receipt.max_fee > 0
    assert receipt.gas_price == 1499999989


@pytest.mark.parametrize(
    "blob_gas_used,blob_gas_key",
    [
        ("0x20000", "blobGasUsed"),
        (131072, "blob_gas_used"),
        (0, "blobGasUsed"),
        (None, "blobGasUsed"),
    ],
)
def test_decode_receipt_shared_blob(ethereum, blob_gas_used, blob_gas_key):
    blob_gas_price = "0x4d137e31b"

    data = {
        "required_confirmations": 0,
        "blockHash": HexBytes("0x051fb508617fcbe0034538b7ee54c1dedbbbaa6f8d0aeb776edf81bafbc883bd"),
        "blockNumber": 10526428,
        "from": "0x194E22F49BC3f58903866d55488E1e9e8d69b517",
        "gas": 5500000,
        "gasPrice": 1000000008,
        "maxFeePerGas": 150000000000,
        "maxPriorityFeePerGas": 1000000000,
        "maxFeePerBlobGas": "0x22ecb25c00",
        "hash": HexBytes("0xd2882bae0d79a6c8e0fbf0089bbcb4b2eef3a1365471ad9f779b06a41ba47d3c"),
        "input": HexBytes("0xb72d42a1000000000000000000000000000000000000000000"),  # Note: abridged
        "nonce": 329925,
        "to": "0xd5c325D183C592C94998000C5e0EED9e6655c020",
        "transactionIndex": 48,
        "value": 0,
        "type": 3,
        "accessList": [],
        "chainId": 5,
        "blobVersionedHashes": [
            "0x015405d502a27897ad38ffcb80566d1559fa53764befc6d90731082837c2d19e"
        ],
        "v": 0,
        "r": HexBytes("0x4e5bdcbba31548cb8f9806491c5ced7b0f1eac78c7dc0677616c23ee0e469f74"),
        "s": HexBytes("0x31c98ea044c97225d83b9a3f0fa719e2299b9fbc6436e4b3eb096ea528de03ff"),
        "yParity": 0,
        "blobGasPrice": blob_gas_price,
        blob_gas_key: blob_gas_used,
        "contractAddress": None,
        "cumulativeGasUsed": 23085827,
        "effectiveGasPrice": 1000000008,
        "gasUsed": 219756,
        "logs": [],
        "logsBloom": HexBytes("0x00000000010002"),
        "status": 1,
        "transactionHash": HexBytes(
            "0xd2882bae0d79a6c8e0fbf0089bbcb4b2eef3a1365471ad9f779b06a41ba47d3c"
        ),
    }
    actual = ethereum.decode_receipt(data)
    assert isinstance(actual, SharedBlobReceipt)
    assert actual.blob_gas_price == int(blob_gas_price, 16)

    if blob_gas_used:
        # all test-values are this when they exist.
        assert actual.blob_gas_used == 131072
    else:
        # when None, should also default to 0.
        assert actual.blob_gas_used == 0

    # Show type=3 is required.
    data["type"] = 2
    actual = ethereum.decode_receipt(data)
    assert not isinstance(actual, SharedBlobReceipt)
    assert isinstance(actual, Receipt)


def test_decode_receipt_misleading_blob_receipt(ethereum):
    """
    Tests a strange situation (noticed on Tenderly nodes) where _some_
    of the keys indicate blob-related fields, set to ``0``, and others
    are missing, because it's not actually a blob receipt. In this case,
    don't use the blob-receipt class.
    """
    data = {
        "type": 2,
        "status": 1,
        "cumulativeGasUsed": 10565720,
        "logsBloom": HexBytes(
            "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        ),
        "logs": [],
        "transactionHash": HexBytes(
            "0x62fc9991bc7fb0c76bc83faaa8d1c17fc5efb050542e58ac358932f80aa7a087"
        ),
        "from": "0x1f9090aaE28b8a3dCeaDf281B0F12828e676c326",
        "to": "0xeBec795c9c8bBD61FFc14A6662944748F299cAcf",
        "contractAddress": None,
        "gasUsed": 21055,
        "effectiveGasPrice": 7267406643,
        "blockHash": HexBytes("0xa47fc133f829183b751488c1146f1085451bcccd247db42066dc6c89eaf5ebac"),
        "blockNumber": 21051245,
        "transactionIndex": 130,
        "blobGasUsed": 0,
    }
    actual = ethereum.decode_receipt(data)
    assert not isinstance(actual, SharedBlobReceipt)
    assert isinstance(actual, Receipt)


def test_default_transaction_type_not_connected_used_default_network(project, ethereum, networks):
    value = TransactionType.STATIC.value
    config_dict = {"ethereum": {"mainnet_fork": {"default_transaction_type": value}}}
    assert ethereum.default_transaction_type == TransactionType.DYNAMIC

    with project.temp_config(**config_dict):
        ethereum._default_network = "mainnet-fork"
        provider = networks.active_provider

        # Disconnect so it uses default.
        networks.active_provider = None
        try:
            assert ethereum.default_network_name == "mainnet-fork"
            assert ethereum.default_transaction_type == TransactionType.STATIC
            ethereum._default_network = LOCAL_NETWORK_NAME
        finally:
            networks.active_provider = provider


def test_default_transaction_type_configured_from_local_network(
    eth_tester_provider, ethereum, project
):
    _ = eth_tester_provider  # Connection required so 'ethereum' knows the network.
    value = TransactionType.STATIC.value
    config = {"ethereum": {LOCAL_NETWORK_NAME: {"default_transaction_type": value}}}
    with project.temp_config(**config):
        assert ethereum.default_transaction_type == TransactionType.STATIC


def test_default_transaction_type_changed_at_class_level(ethereum):
    """
    Simulates an L2 plugin changing the default at the definition-level.
    """

    class L2NetworkConfig(BaseEthereumConfig):
        DEFAULT_TRANSACTION_TYPE: ClassVar[int] = TransactionType.STATIC.value

    config = L2NetworkConfig()
    assert config.local.default_transaction_type.value == 0
    assert config.mainnet.default_transaction_type.value == 0
    assert config.mainnet_fork.default_transaction_type.value == 0


@pytest.mark.parametrize("network_name", (LOCAL_NETWORK_NAME, "mainnet-fork", "mainnet_fork"))
def test_gas_limit_local_networks(ethereum, network_name):
    network = ethereum.get_network(network_name)
    network.__dict__.pop("gas_limit", None)  # Refresh in case was changed in another test.
    assert network.gas_limit == "max"


def test_gas_limit_live_networks(ethereum):
    network = ethereum.get_network("sepolia")
    assert network.gas_limit == AutoGasLimit(multiplier=1.0)


def test_encode_blueprint_contract(ethereum, project):
    actual = ethereum.encode_contract_blueprint(project.VyperContract.contract_type)
    ct_bytes = project.VyperContract.contract_type.deployment_bytecode.to_bytes()

    # EIP-5202
    expected = (
        HexBytes("0x61")
        + (len(ct_bytes) + 3).to_bytes(2, "big")  # preface and length
        + HexBytes("0x3d81600a3d39f3")  # return stuff
        + BLUEPRINT_HEADER
        + HexBytes(0)
        + ct_bytes
    )

    assert actual.data == HexBytes(expected)


def test_decode_returndata(ethereum):
    abi = make_method_abi("doThing", outputs=[{"name": "", "type": "bool"}])
    data = HexBytes32.__eth_pydantic_validate__(0)
    actual = ethereum.decode_returndata(abi, data)
    assert actual == (False,)


def test_decode_returndata_non_empty_padding_bytes(ethereum):
    raw_data = HexBytes(
        "0x08c379a000000000000000000000000000000000000000000000000000000000000000200"
        "000000000000000000000000000000000000000000000000000000000000012696e73756666"
        "696369656e742066756e64730000000000000000000000000000"
    )
    abi = make_method_abi(
        "transfer",
        inputs=[
            {"name": "receiver", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        outputs=[{"name": "", "type": "bool"}],
    )
    with pytest.raises(DecodingError):
        ethereum.decode_returndata(abi, raw_data)


def test_decode_returndata_no_bytes_returns_zero(ethereum):
    abi = make_method_abi("doThing", outputs=[{"name": "", "type": "bool"}])
    actual = ethereum.decode_returndata(abi, b"")
    assert actual == (0,)


def test_decode_returndata_list_with_1_struct(ethereum):
    """
    Tests a condition where an array of a list with 1 struct
    would be turned into a raw tuple instead of the Struct class.
    """
    abi = make_method_abi(
        "getArrayOfStructs",
        outputs=[
            ABIType(
                name="",
                type="tuple[]",
                components=[
                    ABIType(name="a", type="address", components=None, internal_type=None),
                    ABIType(name="b", type="bytes32", components=None, internal_type=None),
                ],
                internal_type=None,
            )
        ],
    )
    raw_data = HexBytes(
        "0x000000000000000000000000000000000000000000000000000000000000002"
        "00000000000000000000000000000000000000000000000000000000000000001"
        "000000000000000000000000f39fd6e51aad88f6f4ce6ab8827279cfffb92266f"
        "d91dc47b758f65fd0f6fe511566770c0ae3f94ff9999ceb23bfec3ac9fdc168"
    )
    actual = ethereum.decode_returndata(abi, raw_data)

    assert len(actual) == 1
    # The result should still be a list.
    # There was also a bug where it was mistakenly a tuple!
    assert isinstance(actual[0], list)
    assert len(actual[0]) == 1
    assert actual[0][0].a == "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
    assert actual[0][0].b == HexBytes(
        "0xfd91dc47b758f65fd0f6fe511566770c0ae3f94ff9999ceb23bfec3ac9fdc168"
    )


def test_decode_returndata_returns_str_comparable_ints(ethereum):
    abi = make_method_abi(
        "gimmeInts",
        outputs=[ABIType(name="", type="int", components=None, internal_type="int")],
    )
    raw_data = HexBytes("0x000000000000000000000000000000000000000000000000000000000000002")
    actual = ethereum.decode_returndata(abi, raw_data)
    assert isinstance(actual[0], int)
    assert isinstance(actual[0], CurrencyValueComparable)


def test_decode_returndata_bool(ethereum):
    abi = MethodABI(
        type="function",
        name="view_method",
        stateMutability="view",
        inputs=[],
        outputs=[ABIType(name="", type="bool", components=None, internal_type=None)],
    )
    raw_data = HexBytes("0x000000000000000000000000000000000000000000000000000000000000001")
    actual = ethereum.decode_returndata(abi, raw_data)[0]
    assert isinstance(actual, bool)


@pytest.mark.parametrize("tx_type", TransactionType)
def test_create_transaction_uses_network_gas_limit(tx_type, ethereum, eth_tester_provider, owner):
    tx = ethereum.create_transaction(type=tx_type.value, sender=owner.address)
    assert tx.type == tx_type.value
    assert tx.gas_limit == eth_tester_provider.max_gas


def test_create_transaction_not_connected(networks_disconnected):
    tx = networks_disconnected.ethereum.create_transaction()
    # Show it doesn't try to refer to network gas limit.
    assert tx.gas_limit == 0


def test_create_transaction_with_none_values(ethereum, eth_tester_provider):
    """
    Tests against None being in place of values in kwargs,
    causing their actual defaults not to get used and ValidationErrors
    to occur.
    """
    static = ethereum.create_transaction(
        value=None, data=None, chain_id=None, gas_price=None, nonce=None, receiver=None
    )
    dynamic = ethereum.create_transaction(
        value=None,
        data=None,
        chain_id=None,
        max_fee=None,
        max_priority_fee=None,
        nonce=None,
        receiver=None,
    )
    assert isinstance(static, StaticFeeTransaction)  # Because used gas_price
    assert isinstance(dynamic, DynamicFeeTransaction)  # Because used max_fee

    for tx in (static, dynamic):
        assert tx.data == b""  # None is not allowed.
        assert tx.value == 0  # None is same as 0.
        assert tx.chain_id == eth_tester_provider.chain_id
        assert tx.nonce is None

    assert static.gas_price is None
    assert dynamic.max_fee is None
    assert dynamic.max_priority_fee is None


@pytest.mark.parametrize("kwarg_name", ("max_fee", "max_fee_per_gas", "maxFee", "maxFeePerGas"))
def test_create_transaction_max_fee_per_gas(kwarg_name, ethereum):
    fee = 1_000_000_000
    tx = ethereum.create_transaction(**{kwarg_name: fee})
    assert tx.max_fee == fee


@pytest.mark.parametrize("kwarg_name", ("gas_price", "gasPrice"))
def test_create_transaction_with_gas_price(kwarg_name, ethereum):
    price = 123
    tx = ethereum.create_transaction(**{kwarg_name: price})
    assert tx.gas_price == price


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 3}, {"max_fee_per_blob_gas": 123}, {"blob_versioned_hashes": [HexBytes(123)]}],
)
def test_create_transaction_shared_blob(tx_kwargs, ethereum):
    tx = ethereum.create_transaction(**tx_kwargs)
    assert isinstance(tx, SharedBlobTransaction)
    assert tx.type == TransactionType.SHARED_BLOB.value


@pytest.mark.parametrize(
    "kwarg_name,value", [("max_fee_per_blob_gas", 345), ("maxFeePerBlobGas", "0x2343")]
)
def test_create_transaction_max_fee_per_blob_gas(kwarg_name, value, ethereum):
    tx = ethereum.create_transaction(**{kwarg_name: value})
    assert isinstance(tx, SharedBlobTransaction)
    expected = value if isinstance(value, int) else int(value, 16)
    assert tx.max_fee_per_blob_gas == expected


@pytest.mark.parametrize(
    "kwarg_name,value",
    [("blob_versioned_hashes", "0x123"), ("blobVersionedHashes", HexBytes("0x123"))],
)
def test_create_transaction_blob_versioned_hashed(kwarg_name, value, ethereum):
    tx = ethereum.create_transaction(**{kwarg_name: [value]})
    assert isinstance(tx, SharedBlobTransaction)
    assert tx.blob_versioned_hashes == [HexBytes(value)]


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 4}, {"authorizationList": []}],
)
def test_create_transaction_set_code(tx_kwargs, ethereum):
    tx = ethereum.create_transaction(**tx_kwargs)
    assert isinstance(tx, SetCodeTransaction)
    assert tx.type == TransactionType.SET_CODE.value


def test_create_transaction_chain_id(ethereum):
    """
    Tests against a bug where `create_transaction()` always used
    the connected chain ID instead of the given.
    """
    chain_id = 123  # Anything but the local, connected chain.
    tx = ethereum.create_transaction(chain_id=chain_id)
    assert tx.chain_id == chain_id


@pytest.mark.parametrize("tx_type", TransactionType)
def test_encode_transaction(tx_type, ethereum, vyper_contract_instance, owner, eth_tester_provider):
    abi = vyper_contract_instance.contract_type.methods[0]
    actual = ethereum.encode_transaction(
        vyper_contract_instance.address, abi, sender=owner.address, type=tx_type.value
    )
    assert actual.gas_limit == eth_tester_provider.max_gas


def test_set_default_network_not_exists(ethereum):
    bad_network = "NOT_EXISTS"
    expected = f"No network in 'ethereum' named '{bad_network}'. Options:.*"
    with pytest.raises(NetworkNotFoundError, match=expected):
        ethereum.set_default_network(bad_network)


def test_networks(ethereum):
    actual = ethereum.networks

    with_forks = ("sepolia", "mainnet")
    without_forks = (LOCAL_NETWORK_NAME,)

    def assert_net(
        net: str,
    ):
        assert net in actual
        base_class = ForkedNetworkAPI if net.endswith("-fork") else NetworkAPI
        assert isinstance(actual[net], base_class)

    for net in with_forks:
        assert_net(net)
        assert_net(f"{net}-fork")
    for net in without_forks:
        assert_net(net)


def test_networks_includes_custom_networks(
    ethereum, custom_networks_config_dict, project, custom_network_name_0, custom_network_name_1
):
    with project.temp_config(**custom_networks_config_dict):
        actual = ethereum.networks
        for net in (
            "sepolia",
            "mainnet",
            LOCAL_NETWORK_NAME,
            custom_network_name_0,
            custom_network_name_1,
            # Show custom networks are auto-forked!
            f"{custom_network_name_0}-fork",
            f"{custom_network_name_1}-fork",
        ):
            assert net in actual
            assert isinstance(actual[net], NetworkAPI)


def test_networks_when_custom_ecosystem(
    configured_custom_ecosystem, networks, custom_network_name_0
):
    obj = networks.custom_ecosystem
    actual = obj.networks
    assert obj.name == CUSTOM_ECOSYSTEM_NAME
    assert custom_network_name_0 in actual
    assert "mainnet" not in actual


def test_networks_multiple_networks_with_same_name(custom_networks_config_dict, ethereum, project):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["name"] = "foonet"
    data["networks"]["custom"][1]["name"] = "foonet"
    expected = ".*More than one network named 'foonet' in ecosystem 'ethereum'.*"
    with project.temp_config(**data):
        with pytest.raises(NetworkError, match=expected):
            _ = ethereum.networks


def test_getattr(ethereum):
    assert ethereum.mainnet.name == "mainnet"
    assert isinstance(ethereum.mainnet, NetworkAPI)


def test_getattr_custom_networks(
    ethereum, custom_networks_config_dict, project, custom_network_name_0
):
    with project.temp_config(**custom_networks_config_dict):
        actual = getattr(ethereum, custom_network_name_0)
        assert actual.name == custom_network_name_0
        assert isinstance(actual, NetworkAPI)


def test_default_network(ethereum):
    assert ethereum.default_network_name == LOCAL_NETWORK_NAME


def test_default_network_when_custom_and_set_in_config(
    configured_custom_ecosystem, networks, custom_network_name_0
):
    ecosystem = networks.get_ecosystem(CUSTOM_ECOSYSTEM_NAME)
    # Force it to use config value (in case was set from previous test)
    ecosystem._default_network = None
    assert ecosystem.default_network_name == custom_network_name_0


def test_default_network_name_set_programmatically(ethereum):
    ethereum._default_network = "testnet"
    assert ethereum.default_network_name == "testnet"
    ethereum._default_network = None


def test_default_network_name_from_config(project, ethereum):
    cfg = {"ethereum": {"default_network": "sepolia"}}
    ethereum._default_network = None
    with project.temp_config(**cfg):
        assert ethereum.default_network_name == "sepolia"


def test_default_network_name_when_not_set_uses_local(project, ethereum):
    orig = project.config.ethereum
    data = orig if isinstance(orig, dict) else orig.model_dump()
    data = {k: v for k, v in data.items() if k not in ("default_network",)}
    data["default_network"] = None
    with project.temp_config(**data):
        ethereum._default_network = None
        assert ethereum.default_network_name == LOCAL_NETWORK_NAME


def test_default_network_name_when_not_set_and_no_local_uses_only(
    project, custom_networks_config_dict
):
    """
    Tests a condition that is rare but when a default network is
    not set but there is a single network. In this case, it
    should use the single network by default.
    """
    net = copy.deepcopy(custom_networks_config_dict["networks"]["custom"][0])

    # In a situation with an ecosystem with only a single network.
    ecosystem_name = "acustomeco"
    net["ecosystem"] = ecosystem_name

    only_network = "onlynet"  # More obvious name for test.
    net["name"] = only_network

    with project.temp_config(networks={"custom": [net]}):
        ecosystem = project.network_manager.get_ecosystem(ecosystem_name)
        ecosystem._default_network = None
        actual = ecosystem.default_network_name

        if actual == LOCAL_NETWORK_NAME:
            # For some reason, this test is flake-y. Offer more info
            # to try and debug when this happens (intermittent CI failure).
            all_nets = ", ".join([x for x in ecosystem.networks.keys()])
            pytest.fail(
                f"assert '{LOCAL_NETWORK_NAME}' == '{only_network}'. More info below:\n"
                f"ecosystem_name={ecosystem.name}\n"
                f"networks={all_nets}\n"
                f"type={type(ecosystem)}"
            )
        else:
            # This should pass.
            assert ecosystem.default_network_name == only_network


def test_decode_custom_error(chain, ethereum):
    data = HexBytes("0x6a12f104")
    abi = [ErrorABI(type="error", name="InsufficientETH")]
    contract_type = ContractType(abi=abi)
    addr = cast(AddressType, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")

    # Hack in contract-type.
    chain.contracts[addr] = contract_type

    actual = ethereum.decode_custom_error(data, addr)
    assert isinstance(actual, CustomError)


def test_decode_custom_error_contract_type_not_found(ethereum):
    data = HexBytes("0x6a12f104")
    # not known
    addr = cast(AddressType, "0x4fC92A3afd70395Cd496C647d5a6CC9D4B2b7FAD")
    actual = ethereum.decode_custom_error(data, addr)
    assert actual is None


def test_decode_custom_error_tx_unsigned(ethereum):
    data = HexBytes("0x6a12f104")
    # not known
    addr = cast(AddressType, "0x4fC92A3afd70395Cd496C647d5a6CC9D4B2b7FAD")
    actual = ethereum.decode_custom_error(data, addr)
    assert actual is None


def test_decode_custom_error_selector_not_found(chain, ethereum):
    data = HexBytes("0x6a12f104")
    abi: list = []
    contract_type = ContractType(abi=abi)
    addr = cast(AddressType, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")

    # Hack in contract-type.
    chain.contracts.contract_types[addr] = contract_type

    tx = ethereum.create_transaction()
    actual = ethereum.decode_custom_error(data, addr, txn=tx)
    assert actual is None


def test_enrich_trace(ethereum, vyper_contract_instance, owner):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = TransactionTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert isinstance(actual, TransactionTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))


def test_enrich_trace_handles_call_type_enum(ethereum, vyper_contract_instance, owner):
    """
    Testing a custom trace whose call tree uses an Enum type instead of a str.
    """

    class PluginTxTrace(TransactionTrace):
        def get_calltree(self) -> CallTreeNode:
            call = super().get_calltree()
            # Force enum value instead of str.
            call.call_type = CallType.CALL
            return call

    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = PluginTxTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert isinstance(actual, PluginTxTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))

    # Hook into helper method hackily with an enum.
    # Notice the mode is Python and not JSON here.
    call = trace.get_calltree().model_dump(by_alias=True)
    actual = ethereum._enrich_calltree(call)
    assert actual["call_type"] == CallType.CALL.value


def test_enrich_trace_handles_events(ethereum, vyper_contract_instance, owner):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)

    # Used Hardhat to get the data.
    events = [
        {
            "call_type": "EVENT",
            "data": "0x3ee0dda47a082585c3c66434c4b5b1b4558581efb5f4ee60a59a58bc1201404b00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000060000000000000000000000000000000000000000000000000000000000000000744796e616d696300000000000000000000000000000000000000000000000000",
            "depth": 1,
            "topics": [
                "0xa84473122c11e32cd505595f246a28418b8ecd6cf819f4e3915363fad1b8f968",
                "0x000000000000000000000000000000000000000000000000000000000000007b",
                "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d",
            ],
        }
    ]

    calldata = "0x3fb5c1cb000000000000000000000000000000000000000000000000000000000000007b"
    call = {
        "events": events,
        "call_type": "CALL",
        "address": vyper_contract_instance.address,
        "calldata": calldata,
    }

    class MyTrace(TransactionTrace):
        def get_calltree(self) -> CallTreeNode:
            return CallTreeNode.model_validate(call)

    trace = MyTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert "events" in actual._enriched_calltree, "is evm-trace updated?"
    events = actual._enriched_calltree["events"]
    expected = [
        {
            "name": "NumberChange",
            "calldata": {
                "b": "0x3ee0..404b",
                "prevNum": 0,
                "dynData": '"Dynamic"',
                "newNum": 123,
                "dynIndexed": "0x9f3d..a94d",
            },
        }
    ]
    assert events == expected


def test_enrich_trace_avoids_using_wrong_contract_type(
    ethereum, vyper_contract_instance, solidity_contract_instance, owner, chain
):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = TransactionTrace(transaction_hash=tx.txn_hash)

    # Purposely passing in the wrong contract type here.
    actual = ethereum.enrich_trace(trace, contract_type=solidity_contract_instance.contract_type)

    assert isinstance(actual, TransactionTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))

    # Ensure the contract type was not corrupted.
    ct = chain.contracts[vyper_contract_instance.address]
    assert ct == vyper_contract_instance.contract_type


def test_get_deployment_address(ethereum, owner, project):
    actual = ethereum.get_deployment_address(owner.address, owner.nonce)
    expected = owner.deploy(project.VyperContract, 490)
    assert actual == expected



import pytest
from eth_pydantic_types import HexBytes
from eth_utils import to_hex
from web3.types import BlockData

from ape_ethereum.ecosystem import Block


@pytest.fixture
def block(chain):
    return chain.blocks.head


def test_block(eth_tester_provider, vyper_contract_instance):
    data = eth_tester_provider.web3.eth.get_block("latest")
    actual = Block.model_validate(data)
    assert actual.hash == data["hash"]
    assert actual.number == data["number"]


def test_repr(block):
    actual = repr(block)
    expected = f"<Block number={block.number} hash={to_hex(block.hash)}>"
    assert actual == expected

    # Show it works when there is no hash.
    block.hash = None
    actual = repr(block)
    expected = f"<Block number={block.number}>"
    assert actual == expected


@pytest.mark.parametrize("mode", ("json", "python"))
def test_model_dump(block, mode):
    actual = block.model_dump(mode=mode)
    expected = {
        "baseFeePerGas": 1000000000,
        "difficulty": 0,
        "gasLimit": 30029122,
        "gasUsed": 0,
        "hash": to_hex(block.hash),
        "num_transactions": 0,
        "number": 0,
        "parentHash": to_hex(block.parent_hash),
        "size": block.size,
        "timestamp": block.timestamp,
        "totalDifficulty": 0,
        "transactions": [],
        "uncles": [],
    }
    assert actual == expected


def test_model_dump_json(block):
    actual = block.model_dump_json()
    expected = (
        '{"baseFeePerGas":1000000000,"difficulty":0,"gasLimit":30029122,"gasUsed":0,'
        f'"hash":"{to_hex(block.hash)}",'
        '"num_transactions":0,"number":0,'
        f'"parentHash":"{to_hex(block.parent_hash)}",'
        f'"size":{block.size},"timestamp":{block.timestamp},'
        '"totalDifficulty":0,"transactions":[],"uncles":[]}'
    )
    assert actual == expected


@pytest.mark.parametrize("size", (123, HexBytes(123), to_hex(123)))
def test_size(block, size):
    block._size = size
    dictionary_python = block.model_dump(mode="python")
    dictionary_json = block.model_dump(mode="json")
    jons_str = block.model_dump_json()
    assert dictionary_python["size"] == 123
    assert dictionary_json["size"] == 123
    assert '"size":123' in jons_str

    # Show the same when validated with size in the model.
    data = block.model_dump()
    data["size"] = size
    new_block = Block.model_validate(data)
    dictionary_python = new_block.model_dump(mode="python")
    dictionary_json = new_block.model_dump(mode="json")
    jons_str = new_block.model_dump_json()
    assert dictionary_python["size"] == 123
    assert dictionary_json["size"] == 123
    assert '"size":123' in jons_str


def test_block_calculate_size(block):
    original = block.model_dump(by_alias=True)
    size = original.pop("size")

    # Show size works normally (validated when passed in as a field).
    assert size > 0
    assert block.size == size

    # Show we can also calculate size if it is missing.
    actual = block.model_validate(original)  # re-init without size.
    assert actual.size == size

    original["size"] = 123
    new_block = Block.model_validate(original)
    assert new_block.size == 123  # Show no clashing.
    assert actual.size == size  # Show this hasn't changed.


def test_block_uncles(block):
    data = block.model_dump(by_alias=True)
    uncles = [HexBytes("0xb983ecae1ed260dd08d108653912a9138bdce56c78aa7d78ee4fca70c2c8767b")]
    data["uncles"] = uncles
    actual = Block.model_validate(data)
    assert actual.uncles == uncles


def test_model_dump_and_validate(block):
    model_dump = block.model_dump(by_alias=True)
    model_validate = Block.model_validate(model_dump)
    assert model_validate == block
    # Validate existing model.
    model_validate_from_model = Block.model_validate(block)
    assert model_validate_from_model == block


def test_model_validate_web3_block():
    """
    Show we have good compatibility with web3.py native types.
    """
    data = BlockData(number=123, timestamp=123, gasLimit=123, gasUsed=100)  # type: ignore
    actual = Block.model_validate(data)
    assert actual.number == 123


def test_transactions(block):
    actual = block.transactions
    expected: list = []
    assert actual == expected

    # Ensure still works when hash is None (was a bug where this crashed).
    block.hash = None
    block.__dict__.pop("transactions", None)  # Ensure not cached.
    assert block.transactions == []



from os import environ

import pytest
from cchecksum import to_checksum_address
from eip712.messages import EIP712Domain, EIP712Message
from eth_account.messages import encode_defunct
from eth_pydantic_types import HexBytes, abi
from eth_utils import keccak, to_hex
from ethpm_types import ContractType
from pydantic import BaseModel

import ape
from ape.api import ImpersonatedAccount
from ape.contracts import ContractContainer
from ape.exceptions import (
    AccountsError,
    AliasAlreadyInUseError,
    MethodNonPayableError,
    MissingDeploymentBytecodeError,
    NetworkError,
    ProjectError,
    SignatureError,
)
from ape.types.gas import AutoGasLimit
from ape.types.signatures import recover_signer
from ape.utils.testing import DEFAULT_TEST_MNEMONIC
from ape_accounts.accounts import (
    ApeSigner,
    KeyfileAccount,
    _get_signing_message_with_display,
    generate_account,
    import_account_from_mnemonic,
    import_account_from_private_key,
)
from ape_ethereum.ecosystem import ProxyType
from ape_ethereum.transactions import TransactionType
from ape_test.accounts import TestAccount
from tests.conftest import explorer_test

# NOTE: Even though `__test__` is set to `False` on the class,
# it must be set here as well for it to properly work for some reason.
TestAccount.__test__ = False

MISSING_VALUE_TRANSFER_ERR_MSG = "Must provide 'VALUE' or use 'send_everything=True"

APE_TEST_PATH = "ape_test.accounts.TestAccount"
APE_ACCOUNTS_PATH = "ape_accounts.accounts.KeyfileAccount"

PASSPHRASE = "asdf1234"
INVALID_PASSPHRASE = "incorrect passphrase"
MNEMONIC = "test test test test test test test test test test test junk"
PRIVATE_KEY = "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"


@pytest.fixture(params=(APE_TEST_PATH, APE_ACCOUNTS_PATH))
def core_account(request, owner, keyfile_account):
    if request.param == APE_TEST_PATH:
        yield owner  # from ape_test plugin

    elif request.param == APE_ACCOUNTS_PATH:
        yield keyfile_account  # from ape_accounts plugin


@pytest.fixture
def message():
    return encode_defunct(text="Hello Apes!")


class Baz(BaseModel):
    addr: abi.address


class Foo(EIP712Message):
    eip712_domain = EIP712Domain(name="Foo")

    bar: abi.address
    baz: Baz


def test_sign_message(signer, message):
    signature = signer.sign_message(message)
    assert signer.check_signature(message, signature)


def test_sign_transaction(signer, message, ethereum):
    transaction = ethereum.create_transaction(nonce=0, max_fee=0, max_priority_fee=0)
    signed_transaction = signer.sign_transaction(transaction)
    assert signed_transaction.signature is not None


def test_sign_transaction_using_keyfile_account(keyfile_account, message, ethereum, runner):
    transaction = ethereum.create_transaction(
        nonce=0, max_fee=0, max_priority_fee=0, data="0x21314135413451"
    )

    with runner.isolation(f"y\n{PASSPHRASE}\ny"):
        signed_transaction = keyfile_account.sign_transaction(transaction)

    assert signed_transaction.signature is not None


def test_sign_string(signer):
    message = "Hello Apes!"
    signature = signer.sign_message(message)
    assert signer.check_signature(message, signature)


def test_sign_int(signer):
    message = 4
    signature = signer.sign_message(message)
    assert signer.check_signature(message, signature)


def test_sign_message_unsupported_type_returns_none(signer):
    message = 1234.123
    signature = signer.sign_message(message)
    assert signature is None


def test_recover_signer(signer, message):
    signature = signer.sign_message(message)
    assert recover_signer(message, signature) == signer


def test_sign_eip712_message(signer):
    baz = Baz(addr=signer.address)
    foo = Foo(bar=signer.address, baz=baz)
    signature = signer.sign_message(foo)
    assert signer.check_signature(foo, signature)


def test_sign_message_with_prompts(runner, keyfile_account, message):
    # "y\na\ny": yes sign, password, yes keep unlocked
    start_nonce = keyfile_account.nonce
    with runner.isolation(input=f"y\n{PASSPHRASE}\ny"):
        signature = keyfile_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)

    # # "n": don't sign
    with runner.isolation(input="n\n"):
        signature = keyfile_account.sign_message(message)
        assert signature is None

    # Nonce should not change from signing messages.
    end_nonce = keyfile_account.nonce
    assert start_nonce == end_nonce


def test_sign_eip712_message_shows_custom_types(signer):
    baz = Baz(addr=signer.address)
    foo = Foo(bar=signer.address, baz=baz)
    display_msg, msg = _get_signing_message_with_display(foo)
    expected = """
Signing EIP712 Message
Domain
  Name: Foo
Message
  bar: 0x9965507D1a55bcC2695C58ba16FB37d819B0A4dc
  baz:
    addr: 0x9965507D1a55bcC2695C58ba16FB37d819B0A4dc
""".strip()
    assert display_msg
    assert display_msg.strip() == expected


def test_sign_raw_hash(runner, keyfile_account):
    # NOTE: `message` is a 32 byte raw hash, which is treated specially
    message = b"\xab" * 32

    # "y\na\ny": yes sign raw hash, password, yes keep unlocked
    with runner.isolation(input=f"y\n{PASSPHRASE}\ny"):
        signature = keyfile_account.sign_raw_msghash(message)
        assert keyfile_account.check_signature(message, signature, recover_using_eip191=False)

    # "n\nn": no sign raw hash: don't sign
    with runner.isolation(input="n"):
        signature = keyfile_account.sign_message(message)
        assert signature is None


def test_transfer(sender, receiver, eth_tester_provider, convert):
    initial_receiver_balance = receiver.balance
    initial_sender_balance = sender.balance
    value_str = "24 gwei"
    value_int = convert(value_str, int)

    receipt = sender.transfer(receiver, value_str)

    # Ensure each account balance was affected accordingly
    expected_receiver_balance = initial_receiver_balance + value_int
    expected_sender_loss = receipt.total_fees_paid + value_int
    expected_sender_balance = initial_sender_balance - expected_sender_loss
    assert receiver.balance == expected_receiver_balance
    assert sender.balance == expected_sender_balance, (
        f"difference: {abs(sender.balance - expected_sender_balance)}"
    )


def test_transfer_with_negative_value(sender, receiver):
    with pytest.raises(AccountsError, match="Value cannot be negative."):
        sender.transfer(receiver, value=-1)


def test_transfer_without_value(sender, receiver):
    with pytest.raises(AccountsError, match=MISSING_VALUE_TRANSFER_ERR_MSG):
        sender.transfer(receiver)


def test_transfer_without_value_send_everything_false(sender, receiver):
    with pytest.raises(AccountsError, match=MISSING_VALUE_TRANSFER_ERR_MSG):
        sender.transfer(receiver, send_everything=False)


def test_transfer_without_value_send_everything_true_with_low_gas(sender, receiver):
    initial_receiver_balance = receiver.balance
    initial_sender_balance = sender.balance

    # Clear balance of sender.
    # Use small gas so for sure runs out of money.
    receipt = sender.transfer(receiver, send_everything=True, gas=22000)

    value_given = receipt.value
    total_spent = value_given + receipt.total_fees_paid
    assert sender.balance == initial_sender_balance - total_spent
    assert receiver.balance == initial_receiver_balance + value_given

    expected_err_regex = r"Sender does not have enough to cover transaction value and gas: \d*"
    with pytest.raises(AccountsError, match=expected_err_regex):
        sender.transfer(receiver, send_everything=True)


def test_transfer_without_value_send_everything_true_with_high_gas(sender, receiver):
    initial_receiver_balance = receiver.balance
    initial_sender_balance = sender.balance

    # The gas selected here is very high compared to what actually gets used.
    gas = 25000000

    # Clear balance of sender
    receipt = sender.transfer(receiver, send_everything=True, gas=gas)

    value_given = receipt.value
    total_spent = value_given + receipt.total_fees_paid
    assert sender.balance == initial_sender_balance - total_spent
    assert receiver.balance == initial_receiver_balance + value_given

    # The sender is able to transfer again because they have so much left over
    # from safely using such a high gas before.
    # Use smaller (more expected) amount of gas this time.
    sender.transfer(receiver, send_everything=True, gas=21000)


def test_transfer_with_value_send_everything_true(sender, receiver):
    with pytest.raises(AccountsError, match="Cannot use 'send_everything=True' with 'VALUE'."):
        sender.transfer(receiver, 1, send_everything=True)


def test_transfer_with_prompts(runner, receiver, keyfile_account):
    # "y\na\ny": yes sign, password, yes keep unlocked
    with runner.isolation(f"y\n{PASSPHRASE}\ny"):
        receipt = keyfile_account.transfer(receiver, "1 gwei")
        assert receipt.receiver == receiver

    # "n": don't sign
    with runner.isolation(input="n\n"):
        with pytest.raises(SignatureError):
            keyfile_account.transfer(receiver, "1 gwei")


def test_transfer_using_type_0(sender, receiver, convert):
    initial_balance = receiver.balance
    sender.transfer(receiver, "1 gwei", type=0)
    expected = initial_balance + convert("1 gwei", int)
    assert receiver.balance == expected


def test_transfer_value_of_0(sender, receiver):
    """
    There was a bug where this failed, thinking there was no value.
    """
    initial_balance = receiver.balance
    sender.transfer(receiver, 0)
    assert receiver.balance == initial_balance

    # Also show conversion works.
    sender.transfer(receiver, "0 wei")
    assert receiver.balance == initial_balance


def test_transfer_mixed_up_sender_and_value(sender, receiver):
    """
    Testing the case where the user mixes up the argument order,
    it should show a nicer error than it was previously, as this is
    a common and easy mistake.
    """
    expected = (
        r"Cannot use integer-type for the `receiver` "
        r"argument in the `\.transfer\(\)` method \(this "
        r"protects against accidentally passing the "
        r"`value` as the `receiver`\)."
    )
    with pytest.raises(AccountsError, match=expected):
        sender.transfer(123, receiver)

    # Similarly show using currency-str (may fail for different error).
    expected = r"Invalid `receiver` value: '123 wei'\."
    with pytest.raises(AccountsError, match=expected):
        sender.transfer("123 wei", receiver)


def test_transfer_sign_is_false(sender, receiver):
    with pytest.raises(SignatureError):
        sender.transfer(receiver, "1 gwei", sign=False)


def test_deploy(owner, contract_container, clean_contract_caches):
    contract = owner.deploy(contract_container, 0)
    assert contract.address
    assert contract.txn_hash

    # Deploy again to prove that we get the correct txn_hash below
    owner.deploy(contract_container, 0)

    # Verify can reload same contract from cache
    contract_from_cache = ape.Contract(contract.address)
    assert contract_from_cache.contract_type == contract.contract_type
    assert contract_from_cache.address == contract.address
    assert contract_from_cache.txn_hash == contract.txn_hash


@explorer_test
def test_deploy_and_publish_local_network(owner, contract_container):
    with pytest.raises(ProjectError, match="Can only publish deployments on a live network"):
        owner.deploy(contract_container, 0, publish=True)


@explorer_test
def test_deploy_and_publish_live_network_no_explorer(owner, contract_container, dummy_live_network):
    dummy_live_network.__dict__["explorer"] = None
    expected_message = "Unable to publish contract - no explorer plugin installed."
    with pytest.raises(NetworkError, match=expected_message):
        owner.deploy(contract_container, 0, publish=True, required_confirmations=0)


@explorer_test
def test_deploy_and_publish(
    owner, contract_container, dummy_live_network_with_explorer, mock_explorer
):
    contract = owner.deploy(contract_container, 0, publish=True, required_confirmations=0)
    mock_explorer.publish_contract.assert_called_once_with(contract.address)


@explorer_test
def test_deploy_and_not_publish(
    owner, contract_container, dummy_live_network_with_explorer, mock_explorer
):
    owner.deploy(contract_container, 0, publish=True, required_confirmations=0)
    assert not mock_explorer.call_count


def test_deploy_proxy(owner, vyper_contract_instance, project, chain):
    target = vyper_contract_instance.address
    proxy = owner.deploy(project.SimpleProxy, target)

    # Ensure we can call both proxy and target methods on it.
    assert proxy.implementation  # No attr err
    assert proxy.myNumber  # No attr err

    # Ensure was properly cached.
    assert proxy.address in chain.contracts.contract_types
    assert proxy.address in chain.contracts.proxy_infos

    # Show the cached proxy info is correct.
    proxy_info = chain.contracts.proxy_infos[proxy.address]
    assert proxy_info.target == target
    assert proxy_info.type == ProxyType.Delegate
    assert proxy_info.abi.name == "implementation"

    # Show we get the implementation contract type using the proxy address
    re_contract = chain.contracts.instance_at(proxy.address)
    assert re_contract.contract_type == proxy.contract_type

    # Show proxy methods are not available on target alone.
    target = chain.contracts.instance_at(proxy_info.target)
    assert target.myNumber  # No attr err
    with pytest.raises(AttributeError):
        _ = target.implementation


def test_deploy_instance(owner, vyper_contract_instance):
    """
    Tests against a confusing scenario where you would get a SignatureError when
    trying to deploy a ContractInstance because Ape would attempt to create a tx
    by calling the contract's default handler.
    """

    expected = (
        r"contract argument must be a ContractContainer type, "
        r"such as 'project\.MyContract' where 'MyContract' is the "
        r"name of a contract in your project\."
    )
    with pytest.raises(TypeError, match=expected):
        owner.deploy(vyper_contract_instance)


@pytest.mark.parametrize("bytecode", (None, {}, {"bytecode": "0x"}))
def test_deploy_no_deployment_bytecode(owner, bytecode):
    """
    https://github.com/ApeWorX/ape/issues/1904
    """
    expected = (
        r"Cannot deploy: contract 'Apes' has no deployment-bytecode\. "
        r"Are you attempting to deploy an interface\?"
    )
    contract_type = ContractType.model_validate(
        {"abi": [], "contractName": "Apes", "deploymentBytecode": bytecode}
    )
    contract = ContractContainer(contract_type)
    with pytest.raises(MissingDeploymentBytecodeError, match=expected):
        owner.deploy(contract)


def test_deploy_contract_type(owner, project, clean_contract_caches):
    contract_type = project.VyperContract.contract_type
    contract = owner.deploy(contract_type, 0)
    assert contract.address
    assert contract.txn_hash


def test_deploy_sending_funds_to_non_payable_constructor(project, owner):
    with pytest.raises(
        MethodNonPayableError,
        match=r"Sending funds to a non-payable constructor\.",
    ):
        owner.deploy(project.SolidityContract, 1, value="1 ether")


def test_send_transaction_with_bad_nonce(sender, receiver):
    # Bump the nonce so we can set one that is too low.
    sender.transfer(receiver, "1 gwei", type=0)

    with pytest.raises(AccountsError, match="Invalid nonce, will not publish."):
        sender.transfer(receiver, "1 gwei", type=0, nonce=0)


def test_send_transaction_without_enough_funds(sender, receiver, eth_tester_provider, convert):
    expected = (
        rf"Transfer value meets or exceeds account balance for account '{sender.address}' .*"
        rf"on chain '{eth_tester_provider.chain_id}' using provider '{eth_tester_provider.name}'\."
        rf"\nAre you using the correct account / chain \/ provider combination\?"
        rf"\n\(transfer_value=\d+, balance=\d+\)\."
    )
    with pytest.raises(AccountsError, match=expected):
        sender.transfer(receiver, "10000000000000 ETH")


def test_send_transaction_without_enough_funds_impersonated_account(
    receiver, accounts, eth_tester_provider, convert
):
    address = "0x4838B106FCe9647Bdf1E7877BF73cE8B0BAD5f97"  # Not a test account!
    impersonated_account = ImpersonatedAccount(raw_address=address)
    accounts._impersonated_accounts[address] = impersonated_account

    # Basically, it failed anywhere else besides the AccountsError you get from not
    # enough balance.
    with pytest.raises(SignatureError):
        impersonated_account.transfer(receiver, "10000000000000 ETH")


def test_send_transaction_sets_defaults(sender, receiver):
    receipt = sender.transfer(receiver, "1 GWEI", gas_limit=None, required_confirmations=None)
    assert receipt.gas_limit > 0
    assert receipt.required_confirmations == 0


def test_account_index_access(accounts):
    account = accounts[0]
    assert account.index == 0
    last_account = accounts[-1]
    assert last_account.index == len(accounts) - 1


def test_accounts_splice_access(accounts):
    alice, bob = accounts[:2]
    assert alice == accounts[0]
    assert bob == accounts[1]
    cat = accounts[-1]
    assert cat == accounts[len(accounts) - 1]
    expected = (len(accounts) // 2) if len(accounts) % 2 == 0 else (len(accounts) // 2 + 1)
    assert len(accounts[::2]) == expected


def test_accounts_address_access(owner, accounts):
    assert accounts[owner.address] == owner


def test_accounts_address_access_conversion_fail(account_manager):
    with pytest.raises(
        KeyError,
        match=(
            r"No account with ID 'FAILS'\. "
            r"Do you have the necessary conversion plugins installed?"
        ),
    ):
        _ = account_manager["FAILS"]


def test_accounts_address_access_not_found(accounts):
    address = "0x1222262222222922222222222222222222222222"
    with pytest.raises(KeyError, match=rf"No account with address '{address}'\."):
        _ = accounts[address]


def test_test_accounts_address_access_conversion_fail(accounts):
    with pytest.raises(KeyError, match=r"No account with ID 'FAILS'"):
        _ = accounts["FAILS"]


def test_test_accounts_address_access_not_found(accounts):
    address = "0x1222262222222922222222222222222222222222"
    with pytest.raises(KeyError, match=rf"No account with address '{address}'\."):
        _ = accounts[address]


def test_accounts_contains(accounts, owner):
    assert owner.address in accounts


def test_autosign_messages(runner, keyfile_account, message):
    keyfile_account.set_autosign(True, passphrase=PASSPHRASE)
    signature = keyfile_account.sign_message(message)
    assert keyfile_account.check_signature(message, signature)

    # Re-enable prompted signing
    keyfile_account.set_autosign(False)
    with runner.isolation(input=f"y\n{PASSPHRASE}\n"):
        signature = keyfile_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)


def test_autosign_transactions(runner, keyfile_account, receiver):
    keyfile_account.set_autosign(True, passphrase=PASSPHRASE)
    assert keyfile_account.transfer(receiver, "1 gwei")

    # Re-enable prompted signing
    keyfile_account.set_autosign(False)
    with runner.isolation(input=f"y\n{PASSPHRASE}\n"):
        assert keyfile_account.transfer(receiver, "1 gwei")


def test_impersonate_not_implemented(accounts, address):
    expected_err_msg = (
        r"Provider 'test' does not support impersonating accounts:\\n"
        rf"No account with address '{address}'\."
    )
    with pytest.raises(KeyError, match=expected_err_msg):
        _ = accounts[address]


def test_impersonated_account_ignores_signature_check_on_txn(accounts, address):
    account = ImpersonatedAccount(raw_address=address)

    # Impersonate hack, since no providers in core actually support it.
    accounts._impersonated_accounts[address] = account
    other_0 = accounts[8]
    other_1 = accounts[9]
    txn = other_0.transfer(other_1, "1 gwei").transaction

    # Hack in fake sender.
    txn.sender = address
    actual = txn.serialize_transaction()

    # Normally, you'd get a signature error here, but since the account is registered
    # as impersonated, ape lets it slide because it knows it won't match.
    assert isinstance(actual, bytes)


def test_contract_as_sender_non_fork_network(contract_instance):
    expected_err_msg = (
        r"Provider 'test' does not support impersonating accounts:\\n"
        rf"No account with address '{contract_instance}'\."
    )
    with pytest.raises(KeyError, match=expected_err_msg):
        contract_instance.setNumber(5, sender=contract_instance)


def test_unlock_with_passphrase_and_sign_message(runner, keyfile_account, message):
    keyfile_account.unlock(passphrase=PASSPHRASE)

    # y: yes, sign (note: unlocking makes the key available but is not the same as autosign).
    with runner.isolation(input="y\n"):
        signature = keyfile_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)


def test_unlock_from_prompt_and_sign_message(runner, keyfile_account, message):
    # a = password
    with runner.isolation(input=f"{PASSPHRASE}\n"):
        keyfile_account.unlock()

    # yes, sign the message
    with runner.isolation(input="y\n"):
        signature = keyfile_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)


def test_unlock_with_passphrase_and_sign_transaction(runner, keyfile_account, receiver):
    keyfile_account.unlock(passphrase=PASSPHRASE)
    # y: yes, sign (note: unlocking makes the key available but is not the same as autosign).
    with runner.isolation(input="y\n"):
        receipt = keyfile_account.transfer(receiver, "1 gwei")
        assert receipt.receiver == receiver


def test_unlock_from_prompt_and_sign_transaction(runner, keyfile_account, receiver):
    # a = password
    with runner.isolation(input=f"{PASSPHRASE}\n"):
        keyfile_account.unlock()

    # yes, sign the transaction
    with runner.isolation(input="y\n"):
        receipt = keyfile_account.transfer(receiver, "1 gwei")
        assert receipt.receiver == receiver


def test_unlock_with_passphrase_from_env_and_sign_message(runner, keyfile_account, message):
    ENV_VARIABLE = f"APE_ACCOUNTS_{keyfile_account.alias}_PASSPHRASE"

    # Set environment variable with passphrase
    environ[ENV_VARIABLE] = PASSPHRASE

    # Unlock using environment variable
    keyfile_account.unlock()

    # Account should be unlocked
    assert not keyfile_account.locked

    # y: yes, sign (note: unlocking makes the key available but is not the same as autosign).
    with runner.isolation(input="y\n"):
        signature = keyfile_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)


def test_unlock_with_wrong_passphrase_from_env(keyfile_account):
    ENV_VARIABLE = f"APE_ACCOUNTS_{keyfile_account.alias}_PASSPHRASE"
    # Set environment variable with passphrase
    environ[ENV_VARIABLE] = INVALID_PASSPHRASE

    # Use pytest.raises to assert that InvalidPasswordError is raised
    with pytest.raises(AccountsError, match="Invalid password"):
        # Unlock using environment variable
        keyfile_account.unlock()

    # Account should be unlocked
    assert keyfile_account.locked


def test_unlock_and_reload(runner, account_manager, keyfile_account, message):
    """
    Tests against a condition where reloading after unlocking
    would not honor unlocked state.
    """
    keyfile_account.unlock(passphrase=PASSPHRASE)
    reloaded_account = account_manager.load(keyfile_account.alias)

    # y: yes, sign (note: unlocking makes the key available but is not the same as autosign).
    with runner.isolation(input="y\n"):
        signature = reloaded_account.sign_message(message)
        assert keyfile_account.check_signature(message, signature)


def test_custom_num_of_test_accounts_config(accounts, project):
    custom_number_of_test_accounts = 25
    test_config = {
        "test": {
            "number_of_accounts": custom_number_of_test_accounts,
        }
    }
    with project.temp_config(**test_config):
        assert len(accounts) == custom_number_of_test_accounts


def test_test_accounts_repr(accounts, config):
    actual = repr(accounts)
    assert config.get_config("test").hd_path in actual

    # Show actual test-account repr (address should be checksummed).
    actual = repr(accounts[0])
    assert accounts[0].address in actual


def test_account_comparison_to_non_account(core_account):
    # Before, would get a ConversionError.
    assert core_account != "foo"


def test_create_account(accounts):
    length_at_start = len(accounts)
    created_account = accounts.generate_test_account()

    assert isinstance(created_account, TestAccount)
    assert created_account.index == length_at_start

    length_at_start = len(accounts)
    second_created_account = accounts.generate_test_account()
    assert len(accounts) == length_at_start + 1

    assert created_account.address != second_created_account.address
    assert second_created_account.index == created_account.index + 1

    # Last index should now refer to the last-created account.
    last_idx_acct = accounts[-1]
    assert last_idx_acct.index == second_created_account.index
    assert last_idx_acct.address == second_created_account.address
    assert last_idx_acct.address != accounts[0].address
    assert last_idx_acct.address != created_account.address


def test_dir(core_account):
    actual = dir(core_account)
    expected = [
        "address",
        "alias",
        "balance",
        "call",
        "delegate",
        "delegate_to",
        "deploy",
        "nonce",
        "prepare_transaction",
        "provider",
        "remove_delegate",
        "set_delegate",
        "sign_authorization",
        "sign_message",
        "sign_transaction",
        "transfer",
    ]
    assert sorted(actual) == sorted(expected)


def test_is_not_contract(owner, keyfile_account):
    assert not owner.is_contract
    assert not keyfile_account.is_contract


def test_using_different_hd_path(accounts, project, eth_tester_provider):
    test_config = {
        "test": {
            "hd_path": "m/44'/60'/0/0",
        }
    }

    old_address = accounts[0].address
    original_settings = eth_tester_provider.settings.model_dump(by_alias=True)
    with project.temp_config(**test_config):
        eth_tester_provider.update_settings(test_config["test"])
        new_address = accounts[0].address

    eth_tester_provider.update_settings(original_settings)
    assert old_address != new_address


def test_mnemonic(accounts):
    actual = accounts.mnemonic
    expected = DEFAULT_TEST_MNEMONIC
    assert actual == expected


def test_mnemonic_setter(accounts):
    original_mnemonic = accounts.mnemonic
    new_mnemonic = "candy maple cake sugar pudding cream honey rich smooth crumble sweet treat"
    original_address = accounts[0].address

    # Change.
    accounts.mnemonic = new_mnemonic
    new_address = accounts[0].address

    # Put back.
    accounts.mnemonic = original_mnemonic

    # Assert.
    assert new_address != original_address


def test_iter_test_accounts(accounts):
    accounts.reset()
    accounts = list(iter(accounts))
    actual = len(accounts)
    expected = len(accounts)
    assert actual == expected


def test_declare(contract_container, sender):
    receipt = sender.declare(contract_container)
    assert not receipt.failed


def test_prepare_transaction(ethereum, sender):
    tx = ethereum.create_transaction()
    prepared_tx = sender.prepare_transaction(tx)
    assert prepared_tx.sender == sender.address
    assert prepared_tx.signature is None


def test_prepare_transaction_sign(sender, ethereum):
    tx = ethereum.create_transaction()
    prepared_tx = sender.prepare_transaction(tx, sign=True)
    assert prepared_tx.sender == sender.address
    assert prepared_tx.signature is not None


@pytest.mark.parametrize("tx_type", (TransactionType.STATIC, TransactionType.DYNAMIC))
def test_prepare_transaction_using_auto_gas(sender, ethereum, tx_type):
    params = (
        ("gas_price",) if tx_type is TransactionType.STATIC else ("max_fee", "max_priority_fee")
    )

    def clear_network_property_cached():
        for field in ("gas_limit", "auto_gas_multiplier"):
            if field in ethereum.local.__dict__:
                del ethereum.local.__dict__[field]

    auto_gas = AutoGasLimit(multiplier=1.0)
    original_limit = ethereum.config.local.gas_limit

    try:
        clear_network_property_cached()
        ethereum.config.local.gas_limit = auto_gas
        assert ethereum.local.gas_limit == auto_gas, "Setup failed - auto gas not set."

        # NOTE: Must create tx _after_ setting network gas value.
        tx = ethereum.create_transaction(type=tx_type)

        # Show tx doesn't have these by default.
        assert tx.nonce is None
        for param in params:
            # Custom fields depending on type.
            assert getattr(tx, param) is None, f"'{param}' unexpectedly set."

        # Gas should NOT yet be estimated, as that happens closer to sending.
        assert tx.gas_limit is None

        # Sets fields.
        tx = sender.prepare_transaction(tx)

        # We expect these fields to have been set.
        assert tx.nonce is not None
        assert tx.gas_limit is not None

        # Show multipliers work. First, reset network to use one (hack).
        gas_smaller = tx.gas_limit

        auto_gas.multiplier = 1.1
        ethereum.config.local.gas_limit = auto_gas
        clear_network_property_cached()
        assert ethereum.local.gas_limit == auto_gas, "Setup failed - auto gas multiplier not set."

        tx2 = ethereum.create_transaction(type=tx_type)
        tx2 = sender.prepare_transaction(tx2)
        gas_bigger = tx2.gas_limit
        assert gas_smaller < gas_bigger

        for param in params:
            assert getattr(tx, param) is not None

    finally:
        ethereum.config.local.gas_limit = original_limit
        clear_network_property_cached()


@pytest.mark.parametrize("tx_type", (TransactionType.STATIC, TransactionType.DYNAMIC))
def test_prepare_transaction_and_call_using_max_gas(tx_type, ethereum, sender, eth_tester_provider):
    tx = ethereum.create_transaction(type=tx_type.value)
    tx = sender.prepare_transaction(tx)
    assert tx.gas_limit == eth_tester_provider.max_gas, "Test setup failed - gas limit unexpected."

    actual = sender.call(tx)
    assert not actual.failed


def test_authorizations_transaction(sender, vyper_contract_instance):
    assert not sender.delegate

    # NOTE: 0x23fd0e40 is method_id for `myNumber()`
    # NOTE: Must call something, since `__default__` raises
    with sender.delegate_to(vyper_contract_instance, data="0x23fd0e40") as delegate:
        assert sender.delegate == vyper_contract_instance
        assert delegate.myNumber() == 0

    sender.remove_delegate()
    assert not sender.delegate


def test_public_key(runner, keyfile_account, owner):
    with runner.isolation(input=f"{PASSPHRASE}\ny\n"):
        pub_key = keyfile_account.public_key
        assert isinstance(keyfile_account.public_key, HexBytes)

        # Prove it is the correct public key by deriving the address.
        derived_address = to_checksum_address(keccak(pub_key)[-20:])
        assert derived_address == keyfile_account.address

    # Also, show the test accounts have access to their public key.
    derived_address = to_checksum_address(keccak(owner.public_key)[-20:])
    assert derived_address == owner.address


def test_load_public_key_from_keyfile(runner, keyfile_account):
    with runner.isolation(input=f"{PASSPHRASE}\ny\n"):
        assert isinstance(keyfile_account.public_key, HexBytes)

        assert (
            to_hex(keyfile_account.public_key)
            == "0x8318535b54105d4a7aae60c08fc45f9687181b4fdfc625bd1a753fa7397fed753547f11ca8696646f2f3acb08e31016afac23e630c5d11f59f61fef57b0d2aa5"  # noqa: E501
        )
        # no need for password when loading from the keyfile
        assert keyfile_account.public_key


def test_generate_account(delete_account_after):
    alias = "gentester"
    with delete_account_after(alias):
        account, mnemonic = generate_account(alias, PASSPHRASE)
        assert len(mnemonic.split(" ")) == 12
        assert isinstance(account, KeyfileAccount)
        assert account.alias == alias
        assert account.locked is True
        account.unlock(PASSPHRASE)
        assert account.locked is False


def test_generate_account_invalid_alias(delete_account_after):
    with pytest.raises(AccountsError, match="Longer aliases cannot be hex strings."):
        generate_account(
            "3fbc0ce3e71421b94f7ff4e753849c540dec9ade57bad60ebbc521adcbcbc024", "asdf1234"
        )

    with pytest.raises(AccountsError, match="Alias must be a str"):
        # Testing an invalid type as arg, so ignoring
        generate_account(b"imma-bytestr", "asdf1234")  # type: ignore

    used_alias = "used"
    with delete_account_after(used_alias):
        generate_account(used_alias, "qwerty1")
        with pytest.raises(AliasAlreadyInUseError):
            generate_account(used_alias, "asdf1234")


def test_generate_account_invalid_passphrase():
    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        generate_account("invalid-passphrase", "")

    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        generate_account("invalid-passphrase", b"bytestring")  # type: ignore


def test_generate_account_insecure_passphrase(delete_account_after):
    short_alias = "shortaccount"
    with delete_account_after(short_alias):
        with pytest.warns(UserWarning, match="short"):
            generate_account(short_alias, "short")

    simple_alias = "simpleaccount"
    with delete_account_after(simple_alias):
        with pytest.warns(UserWarning, match="simple"):
            generate_account(simple_alias, "simple")


def test_import_account_from_mnemonic(delete_account_after):
    alias = "iafmtester"
    with delete_account_after(alias):
        account = import_account_from_mnemonic(alias, PASSPHRASE, MNEMONIC)
        assert isinstance(account, KeyfileAccount)
        assert account.alias == alias
        assert account.address == "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        assert account.locked is True
        account.unlock(PASSPHRASE)
        assert account.locked is False


def test_import_account_from_mnemonic_invalid_alias(delete_account_after):
    with pytest.raises(AccountsError, match="Longer aliases cannot be hex strings."):
        import_account_from_mnemonic(
            "3fbc0ce3e71421b94f7ff4e753849c540dec9ade57bad60ebbc521adcbcbc024", "asdf1234", MNEMONIC
        )

    with pytest.raises(AccountsError, match="Alias must be a str"):
        # Testing an invalid type as arg, so ignoring
        import_account_from_mnemonic(b"imma-bytestr", "asdf1234", MNEMONIC)  # type: ignore

    used_alias = "iamfused"
    with delete_account_after(used_alias):
        import_account_from_mnemonic(used_alias, "qwerty1", MNEMONIC)
        with pytest.raises(AliasAlreadyInUseError):
            import_account_from_mnemonic(used_alias, "asdf1234", MNEMONIC)


def test_import_account_from_mnemonic_invalid_passphrase():
    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        import_account_from_mnemonic("invalid-passphrase", "", MNEMONIC)

    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        import_account_from_mnemonic("invalid-passphrase", b"bytestring", MNEMONIC)  # type: ignore


def test_import_account_from_mnemonic_insecure_passphrase(delete_account_after):
    short_alias = "iafmshortaccount"
    with delete_account_after(short_alias):
        with pytest.warns(UserWarning, match="short"):
            import_account_from_mnemonic(short_alias, "short", MNEMONIC)

    simple_alias = "iafmsimpleaccount"
    with delete_account_after(simple_alias):
        with pytest.warns(UserWarning, match="simple"):
            import_account_from_mnemonic(simple_alias, "simple", MNEMONIC)


def test_import_account_from_private_key(delete_account_after):
    alias = "iafpktester"
    with delete_account_after(alias):
        account = import_account_from_private_key(alias, PASSPHRASE, PRIVATE_KEY)
        assert isinstance(account, KeyfileAccount)
        assert account.alias == alias
        assert account.address == "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        assert account.locked is True
        account.unlock(PASSPHRASE)
        assert account.locked is False


def test_import_account_from_private_key_invalid_alias(delete_account_after):
    with pytest.raises(AccountsError, match="Longer aliases cannot be hex strings."):
        import_account_from_private_key(
            "3fbc0ce3e71421b94f7ff4e753849c540dec9ade57bad60ebbc521adcbcbc024",
            "asdf1234",
            PRIVATE_KEY,
        )

    with pytest.raises(AccountsError, match="Alias must be a str"):
        # Testing an invalid type as arg, so ignoring
        import_account_from_private_key(b"imma-bytestr", "asdf1234", PRIVATE_KEY)  # type: ignore

    used_alias = "iafpkused"
    with delete_account_after(used_alias):
        import_account_from_private_key(used_alias, "qwerty1", PRIVATE_KEY)
        with pytest.raises(AliasAlreadyInUseError):
            import_account_from_private_key(used_alias, "asdf1234", PRIVATE_KEY)


def test_import_account_from_private_key_invalid_passphrase():
    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        import_account_from_private_key("invalid-passphrase", "", PRIVATE_KEY)

    with pytest.raises(AccountsError, match="Account file encryption passphrase must be provided."):
        import_account_from_private_key(
            "invalid-passphrase",
            b"bytestring",  # type: ignore
            PRIVATE_KEY,
        )


def test_import_account_from_private_key_insecure_passphrase(delete_account_after):
    short_alias = "iafpkshortaccount"
    with delete_account_after(short_alias):
        with pytest.warns(UserWarning, match="short"):
            import_account_from_private_key(short_alias, "short", PRIVATE_KEY)

    simple_alias = "iafpksimpleaccount"
    with delete_account_after(simple_alias):
        with pytest.warns(UserWarning, match="simple"):
            import_account_from_private_key(simple_alias, "simple", PRIVATE_KEY)


def test_load(account_manager, keyfile_account):
    account = account_manager.load(keyfile_account.alias)
    assert account == keyfile_account


def test_get_deployment_address(owner, project):
    deployment_address_1 = owner.get_deployment_address()
    deployment_address_2 = owner.get_deployment_address(nonce=owner.nonce + 1)
    instance_1 = owner.deploy(project.VyperContract, 490)
    assert instance_1.address == deployment_address_1
    instance_2 = owner.deploy(project.VyperContract, 490)
    assert instance_2.address == deployment_address_2


def test_repr(account_manager):
    """
    NOTE: __repr__ should be simple and fast!
      Previously, we showed the repr of all the accounts.
      That was a bad idea, as that can be very unnecessarily slow.
      Hence, this test exists to ensure care is taken.
    """
    actual = repr(account_manager)
    assert actual == "<AccountManager>"


def test_call(owner, vyper_contract_instance):
    tx = vyper_contract_instance.setNumber.as_transaction(5991)
    receipt = owner.call(tx)
    assert not receipt.failed


def test_call_sign_false(owner, vyper_contract_instance):
    tx = vyper_contract_instance.setNumber.as_transaction(5991)
    with pytest.raises(SignatureError):
        owner.call(tx, sign=False)


def test_resolve_address(owner, keyfile_account, account_manager, vyper_contract_instance):
    # Test test-account alias input.
    actual = account_manager.resolve_address(owner.alias)
    assert actual == owner.address

    # Test keyfile-account alias input.
    actual = account_manager.resolve_address(keyfile_account.alias)
    assert actual == keyfile_account.address

    # Test address input.
    actual = account_manager.resolve_address(owner.address)
    assert actual == owner.address

    # Test account input.
    actual = account_manager.resolve_address(owner)
    assert actual == owner.address

    # Test contract input.
    actual = account_manager.resolve_address(vyper_contract_instance)
    assert actual == vyper_contract_instance.address

    # Test int input.
    actual = account_manager.resolve_address(int(owner.address, 16))
    assert actual == owner.address

    # Test int input.
    actual = account_manager.resolve_address(HexBytes(owner.address))
    assert actual == owner.address


def test_use_ape_signer(accounts, project):
    """
    Showing you can use the inner Ape account (base class) directly as an ape account in txns.
    """
    signer = ApeSigner(private_key=accounts[5].private_key)
    assert isinstance(signer, ApeSigner)
    contract = project.VyperContract.deploy(1012, sender=signer)
    assert contract.is_contract



import functools
from collections.abc import Callable

from .account import AccountPlugin
from .compiler import CompilerPlugin
from .config import Config
from .converter import ConversionPlugin
from .network import EcosystemPlugin, ExplorerPlugin, NetworkPlugin, ProviderPlugin
from .pluggy_patch import PluginType, hookimpl
from .pluggy_patch import plugin_manager as pluggy_manager
from .project import DependencyPlugin, ProjectPlugin
from .query import QueryPlugin


class PluginError(Exception):
    pass


# Combine all the plugins together via subclassing (merges `hookspec`s)
class AllPluginHooks(
    Config,
    AccountPlugin,
    CompilerPlugin,
    ConversionPlugin,
    DependencyPlugin,
    EcosystemPlugin,
    ExplorerPlugin,
    NetworkPlugin,
    ProjectPlugin,
    ProviderPlugin,
    QueryPlugin,
):
    pass


# All hookspecs are registered
pluggy_manager.add_hookspecs(AllPluginHooks)


def get_hooks(plugin_type):
    return [name for name, method in plugin_type.__dict__.items() if hasattr(method, "ape_spec")]


def register(plugin_type: type[PluginType], **hookimpl_kwargs) -> Callable:
    """
    Register your plugin to ape. You must call this decorator to get your plugins
    included in ape's plugin ecosystem.

    Usage example::

        @plugins.register(plugins.AccountPlugin)  # 'register()' example
        def account_types():
            return AccountContainer, KeyfileAccount

    Args:
        plugin_type (type[:class:`~ape.plugins.pluggy_patch.PluginType`]): The plugin
          type to register.

        hookimpl_kwargs: Return-values required by the plugin type.

    Returns:
        Callable
    """

    # NOTE: we are basically checking that `plugin_type`
    #       is one of the parent classes of `Plugins`
    if not issubclass(AllPluginHooks, plugin_type):
        raise PluginError("Not a valid plugin type to register.")

    def check_hook(plugin_type, hookimpl_kwargs, fn):
        fn = hookimpl(fn, **hookimpl_kwargs)

        if not hasattr(plugin_type, fn.__name__):
            hooks = get_hooks(plugin_type)

            raise PluginError(
                f"Registered function `{fn.__name__}` is not"
                f" a valid hook for {plugin_type.__name__}, must be one of:"
                f" {hooks}."
            )

        return fn

    # NOTE: Get around issue with using `plugin_type` raw in `check_hook`
    return functools.partial(check_hook, plugin_type, hookimpl_kwargs)


__all__ = [
    "register",
]



from decimal import Decimal

import pytest

from ape.exceptions import ConversionError


def test_convert_formatted_float_strings_to_decimal(convert):
    test_strings = [
        "1.000",
        "1.00",
        "1.0",
        "0.1",
        "0.01",
        "0.001",
    ]
    for test_string in test_strings:
        actual = convert(test_string, Decimal)
        assert actual == Decimal(test_string)


def test_convert_badly_formatted_float_strings_to_decimal(convert):
    test_strings = [
        ".1",
        "1.",
        ".",
    ]
    for test_string in test_strings:
        with pytest.raises(
            ConversionError, match=f"No conversion registered to handle '{test_string}'"
        ):
            convert(test_string, Decimal)


def test_convert_int_strings(convert):
    test_strings = [
        "1",
        "10",
        "100",
    ]
    for test_string in test_strings:
        with pytest.raises(
            ConversionError, match=f"No conversion registered to handle '{test_string}'"
        ):
            convert(test_string, Decimal)


def test_convert_alphanumeric_strings(convert):
    test_strings = [
        "a",
        "H",
        "XYZ",
    ]
    for test_string in test_strings:
        with pytest.raises(
            ConversionError, match=f"No conversion registered to handle '{test_string}'"
        ):
            convert(test_string, Decimal)


def test_convert_strings_with_token_names(convert):
    test_strings = [
        "0.999 DAI",
        "10.0 USDC",
    ]
    for test_string in test_strings:
        with pytest.raises(
            ConversionError, match=f"No conversion registered to handle '{test_string}'"
        ):
            convert(test_string, Decimal)


def test_convert_strings_with_ether_alias(convert):
    test_strings = [
        "0 wei",
        "999 gwei",
        "1.0 ether",
    ]
    for test_string in test_strings:
        with pytest.raises(
            ConversionError, match=f"No conversion registered to handle '{test_string}'"
        ):
            convert(test_string, Decimal)



from typing import Any

import pytest
from eth_utils import to_hex
from ethpm_types.abi import EventABI
from hexbytes import HexBytes
from pydantic import BaseModel, Field, ValidationError

from ape.types.address import AddressType
from ape.types.basic import HexInt
from ape.types.events import ContractLog, LogFilter
from ape.types.private_mempool import SimulationReport
from ape.types.units import CurrencyValueComparable
from ape.utils.misc import ZERO_ADDRESS

TXN_HASH = "0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa222222222222222222222222"
BLOCK_HASH = "0x999999998d4f99f68db9999999999da27ed049458b139999999999e910155b99"
BLOCK_NUMBER = 323423
EVENT_NAME = "MyEvent"
LOG_INDEX = 7
TXN_INDEX = 2
RAW_LOG: dict = {
    "block_hash": BLOCK_HASH,
    "block_number": BLOCK_NUMBER,
    "contract_address": ZERO_ADDRESS,
    "event_arguments": {"foo": 0, "bar": 1},
    "log_index": LOG_INDEX,
    "event_name": EVENT_NAME,
    "transaction_hash": TXN_HASH,
    "transaction_index": TXN_INDEX,
}
RAW_EVENT_ABI = """
{
  "anonymous": false,
  "inputs": [
    {
      "indexed": true,
      "name": "oldVersion",
      "type": "address"
    },
    {
      "indexed": true,
      "name": "newVersion",
      "type": "address"
    }
  ],
  "name": "StrategyMigrated",
  "type": "event"
}
"""


@pytest.fixture
def log():
    return ContractLog(**RAW_LOG)


def test_contract_log_serialization(log, zero_address):
    obj = ContractLog.model_validate(log.model_dump())
    assert obj.contract_address == zero_address
    assert obj.block_hash == BLOCK_HASH
    assert obj.block_number == BLOCK_NUMBER
    assert obj.event_name == EVENT_NAME
    assert obj.log_index == 7
    assert obj.transaction_hash == TXN_HASH
    assert obj.transaction_index == TXN_INDEX


def test_contract_log_serialization_with_hex_strings_and_non_checksum_addresses(log, zero_address):
    data = log.model_dump()
    data["log_index"] = to_hex(log.log_index)
    data["transaction_index"] = to_hex(log.transaction_index)
    data["block_number"] = to_hex(log.block_number)
    data["contract_address"] = log.contract_address.lower()

    obj = ContractLog(**data)

    assert obj.contract_address == zero_address
    assert obj.block_hash == BLOCK_HASH
    assert obj.block_number == BLOCK_NUMBER
    assert obj.event_name == EVENT_NAME
    assert obj.log_index == 7
    assert obj.transaction_hash == TXN_HASH
    assert obj.transaction_index == TXN_INDEX


def test_contract_log_str(log):
    obj = ContractLog.model_validate(log.model_dump())
    assert str(obj) == "MyEvent(foo=0 bar=1)"


def test_contract_log_repr(log):
    obj = ContractLog.model_validate(log.model_dump())
    assert repr(obj) == "<MyEvent foo=0 bar=1>"


def test_contract_log_access(log):
    assert "foo" in log
    assert "bar" in log
    assert log.foo == log["foo"] == log.get("foo") == 0
    assert log.bar == log["bar"] == log.get("bar") == 1


def test_contract_log_abi(log):
    # The log fixture is very basic class usage.
    assert log.abi.name == "MyEvent"


def test_contract_log_topics(log):
    actual = log.topics
    expected = ["0x4dbfb68b43dddfa12b51ebe99ab8fded620f9a0ac23142879a4f192a1b7952d2"]
    assert actual == expected


def test_topic_filter_encoding():
    event_abi = EventABI.model_validate_json(RAW_EVENT_ABI)
    log_filter = LogFilter.from_event(
        event=event_abi, search_topics={"newVersion": "0x8c44Cc5c0f5CD2f7f17B9Aca85d456df25a61Ae8"}
    )
    assert log_filter.topic_filter == [
        "0x100b69bb6b504e1252e36b375233158edee64d071b399e2f81473a695fd1b021",
        None,
        "0x0000000000000000000000008c44cc5c0f5cd2f7f17b9aca85d456df25a61ae8",
    ]


def test_log_filter_model_validate_coerces_nones():
    log_filter = LogFilter.model_validate(
        {"addresses": None, "events": None, "topic_filter": None, "selectors": None}
    )
    assert log_filter.addresses == []
    assert log_filter.events == []
    assert log_filter.topic_filter == []
    assert log_filter.selectors == {}


def test_address_type(owner):
    class MyModel(BaseModel):
        addr: AddressType

    # Show str works.
    instance_str = MyModel(addr=owner.address)
    assert instance_str.addr == owner.address

    # Show hex bytes work.
    instance_hex_bytes = MyModel(addr=HexBytes(owner.address))
    assert instance_hex_bytes.addr == owner.address

    # Show raw bytes work.
    instance_bytes = MyModel(addr=bytes.fromhex(owner.address[2:]))
    assert instance_bytes.addr == owner.address

    # Show int works.
    instance_bytes = MyModel(addr=int(owner.address, 16))
    assert instance_bytes.addr == owner.address


class TestHexInt:
    def test_model(self):
        class MyModel(BaseModel):
            ual: HexInt = 0
            ual_optional: HexInt | None = Field(default=None, validate_default=True)

        act = MyModel.model_validate({"ual": "0x123"})
        expected = 291  # Base-10 form of 0x123.
        assert act.ual == expected
        assert act.ual_optional is None

    def test_none(self):
        """
        Was getting unhelpful conversion errors here. We should instead
        let Pydantic fail as it normally does in this situation.
        """

        class MyModel(BaseModel):
            an_int: HexInt

        expected = ".*Input should be a valid integer.*"
        with pytest.raises(ValidationError, match=expected):
            _ = MyModel(an_int=None)


class TestCurrencyValueComparable:
    def test_use_for_int_in_pydantic_model(self):
        value = 100000000000000000000000000000000000000000000

        class MyBasicModel(BaseModel):
            val: int

        model = MyBasicModel.model_validate({"val": CurrencyValueComparable(value)})
        assert model.val == value

        # Ensure serializes.
        dumped = model.model_dump()
        assert dumped["val"] == value

    @pytest.mark.parametrize("mode", ("json", "python"))
    def test_use_in_model_annotation(self, mode):
        value = 100000000000000000000000000000000000000000000

        class MyAnnotatedModel(BaseModel):
            val: CurrencyValueComparable
            val_optional: CurrencyValueComparable | None

        model = MyAnnotatedModel.model_validate({"val": value, "val_optional": value})
        assert isinstance(model.val, CurrencyValueComparable)
        assert model.val == value

        # Show can use currency-comparable
        expected_currency_value = "100000000000000000000000000 ETH"
        assert model.val == expected_currency_value
        assert model.val_optional == expected_currency_value

        # Ensure serializes.
        dumped = model.model_dump(mode=mode)
        assert dumped["val"] == value
        assert dumped["val_optional"] == value

    def test_validate_from_currency_value(self):
        class MyAnnotatedModel(BaseModel):
            val: CurrencyValueComparable
            val_optional: CurrencyValueComparable | None
            val_in_dict: dict[str, Any]

        value = "100000000000000000000000000 ETH"
        expected = 100000000000000000000000000000000000000000000
        data = {
            "val": value,
            "val_optional": value,
            "val_in_dict": {"value": CurrencyValueComparable(expected)},
        }
        model = MyAnnotatedModel.model_validate(data)
        for actual in (model.val, model.val_optional, model.val_in_dict["value"]):
            for ex in (value, expected):
                assert actual == ex

    def test_hashable(self):
        mapping: dict[int, str] = {0: "0", 1: "1"}
        key = CurrencyValueComparable(0)
        assert key in mapping
        assert mapping[key] == "0"


class TestSimulationReport:
    def test_model(self):
        data = {
            "success": True,
            "stateBlock": "0xa",
            "mevGasPrice": "0x399339e8",
            "profit": "0x14537d8d4b310",
            "refundableValue": "0x14537d8d4b310",
            "gasUsed": "0x5a60a",
            "logs": [
                {
                    "txLogs": [
                        {
                            "address": "0x5fbdb2315678afecb367f032d93f642f64180aa3",
                            "topics": [
                                "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
                                "0x0000000000000000000000009fe46736679d2d9a65f0992f2272de9f3c7fa6e0",
                                "0x0000000000000000000000003c44cdddb6a900fa2b585dd299e03d12fa4293bc",
                            ],
                            "data": "0x00000000000000000000000000000000000000000000000000000000000186a0",
                            "blockHash": None,
                            "blockNumber": None,
                            "transactionHash": "0x52a4d4f8e0411d06aa5b9bc6ab0a143716225ca8f16568739f80c22d31b8f450",
                            "transactionIndex": "0x0",
                            "logIndex": "0x0",
                            "removed": False,
                        }
                    ]
                },
                {
                    "txLogs": [
                        {
                            "address": "0xe7f1725e7734ce288f8367e1bb143e90bb3f0512",
                            "topics": [
                                "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
                                "0x0000000000000000000000003c44cdddb6a900fa2b585dd299e03d12fa4293bc",
                                "0x0000000000000000000000009fe46736679d2d9a65f0992f2272de9f3c7fa6e0",
                            ],
                            "data": "0x00000000000000000000000000000000000000000000000000000000000186a0",
                            "blockHash": None,
                            "blockNumber": None,
                            "transactionHash": "0xf678d62c462dc5a63504e1703822dc6740771a8cd4b57bf74f660e50792e3c40",
                            "transactionIndex": "0x1",
                            "logIndex": "0x1",
                            "removed": False,
                        }
                    ]
                },
            ],
        }
        report = SimulationReport.model_validate(data)
        assert report.success

        event = EventABI.model_validate(
            {
                "anonymous": False,
                "inputs": [
                    {"indexed": True, "name": "from", "type": "address"},
                    {"indexed": True, "name": "to", "type": "address"},
                    {"indexed": False, "name": "value", "type": "uint256"},
                ],
                "name": "Transfer",
                "type": "event",
            }
        )

        events = list(report.decode_logs(event))
        assert len(events) == 2
        assert events[0].event_name == "Transfer"
        assert events[1].event_name == "Transfer"



import re
from ast import literal_eval
from collections.abc import Iterator
from functools import cached_property
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING, Any, cast

from eth.exceptions import HeaderNotFound
from eth_pydantic_types import HexBytes
from eth_tester import EthereumTester  # type: ignore
from eth_tester.backends import PyEVMBackend  # type: ignore
from eth_tester.backends.pyevm.main import setup_tester_chain  # type: ignore
from eth_tester.exceptions import TransactionFailed  # type: ignore
from eth_utils import is_0x_prefixed, to_hex
from eth_utils.exceptions import ValidationError
from eth_utils.toolz import merge
from web3 import EthereumTesterProvider, Web3
from web3.exceptions import ContractLogicError as Web3ContractLogicError
from web3.providers.eth_tester.defaults import API_ENDPOINTS, static_return
from web3.types import TxParams

from ape.api.providers import BlockAPI, CallResult, TestProviderAPI
from ape.exceptions import (
    APINotImplementedError,
    ContractLogicError,
    ProviderError,
    ProviderNotConnectedError,
    TransactionError,
    UnknownSnapshotError,
    VirtualMachineError,
)
from ape.logging import logger
from ape.types.address import AddressType
from ape.utils.misc import gas_estimation_error_message
from ape.utils.testing import DEFAULT_TEST_HD_PATH
from ape_ethereum.provider import Web3Provider
from ape_ethereum.trace import TraceApproach, TransactionTrace
from ape_ethereum.transactions import TransactionStatusEnum
from ape_test.config import EthTesterProviderConfig

if TYPE_CHECKING:
    from ape.api.accounts import TestAccountAPI
    from ape.api.trace import TraceAPI
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.types.events import ContractLog, LogFilter
    from ape.types.vm import BlockID, SnapshotID
    from ape_test.config import ApeTestConfig


class ApeEVMBackend(PyEVMBackend):
    """
    A lazier version of PyEVMBackend for the Ape framework.
    """

    def __init__(self, config: "ApeTestConfig", chain_id: int):
        self.config = config
        # Lazily set.
        self._chain = None
        self.chain_id = chain_id

    @property
    def hd_path(self) -> str:
        return (self.config.hd_path or DEFAULT_TEST_HD_PATH).rstrip("/")

    @property
    def balance(self) -> int:
        return self.config.balance

    @property
    def genesis_state(self) -> dict:
        return self.generate_genesis_state(
            mnemonic=self.config.mnemonic,
            overrides={"balance": self.balance},
            num_accounts=self.config.number_of_accounts,
            hd_path=self.hd_path,
        )

    @cached_property
    def _setup_tester_chain(self):
        return setup_tester_chain(
            None,
            self.genesis_state,
            self.config.number_of_accounts,
            None,
            self.config.mnemonic,
            self.config.hd_path,
        )

    @property
    def account_keys(self):
        return self._setup_tester_chain[0]

    @property
    def chain(self):
        if self._chain is None:
            # Initial chain.
            self._chain = self._setup_tester_chain[1]
            # HACK: Make sure PyEVM's chain ID is the same as ours
            self._chain.chain_id = self.chain_id  # type: ignore[attr-defined]

        return self._chain

    @chain.setter
    def chain(self, value):
        self._chain = value  # Changes during snapshot reverting.
        # HACK: Make sure PyEVM's chain ID is the same as ours
        self._chain.chain_id = self.chain_id  # type: ignore[attr-defined]


class ApeTester(EthereumTesterProvider):
    def __init__(self, config: "ApeTestConfig", chain_id: int):
        self.config = config
        self.chain_id = chain_id
        self._backend: ApeEVMBackend | None = None
        self._ethereum_tester: EthereumTester | None = None

    @property
    def ethereum_tester(self) -> EthereumTester:
        if self._ethereum_tester is None:
            self._backend = ApeEVMBackend(self.config, self.chain_id)
            self._ethereum_tester = EthereumTester(self._backend)

        return self._ethereum_tester

    @ethereum_tester.setter
    def ethereum_tester(self, value):
        self._ethereum_tester = value
        self._backend = value.backend

    @property
    def _batching_context(self):  # type: ignore
        # TODO: Figure out correct way; remove this hack

        class MockBatchingContext:
            def get(self, *args, **kwargs):
                return None

        return MockBatchingContext()

    @property
    def backend(self) -> "ApeEVMBackend":
        if self._backend is None:
            self._backend = ApeEVMBackend(self.config, self.chain_id)
            self._ethereum_tester = EthereumTester(self._backend)

        return self._backend

    @backend.setter
    def backend(self, value):
        self._backend = value
        self._ethereum_tester = EthereumTester(self._backend)

    @cached_property
    def api_endpoints(self) -> dict:  # type: ignore
        endpoints = {**API_ENDPOINTS}
        endpoints["eth"] = merge(endpoints["eth"], {"chainId": static_return(self.chain_id)})
        return endpoints


class LocalProvider(TestProviderAPI, Web3Provider):
    _evm_backend: PyEVMBackend | None = None
    _CANNOT_AFFORD_GAS_PATTERN: Pattern = re.compile(
        r"Sender b'[\\*|\w]*' cannot afford txn gas (\d+) with account balance (\d+)"
    )
    _INVALID_NONCE_PATTERN: Pattern = re.compile(
        r"Invalid transaction nonce: Expected (\d+), but got (\d+)"
    )

    @property
    def config(self) -> "ApeTestConfig":  # type: ignore
        return super().config  # type: ignore

    @property
    def evm_backend(self) -> ApeEVMBackend:
        return self.tester.backend

    @cached_property
    def tester(self) -> ApeTester:
        return ApeTester(self.config, self.settings.chain_id)

    @property
    def auto_mine(self) -> bool:
        return self.tester.ethereum_tester.auto_mine_transactions

    @auto_mine.setter
    def auto_mine(self, value: Any) -> None:
        if value is True:
            self.tester.ethereum_tester.enable_auto_mine_transactions()
        elif value is False:
            self.tester.ethereum_tester.disable_auto_mine_transactions()
        else:
            raise TypeError("Expecting bool-value for auto_mine setter.")

    @property
    def max_gas(self) -> int:
        return self.evm_backend.get_block_by_number("latest")["gas_limit"]

    @property
    def http_uri(self) -> str | None:
        return None

    @property
    def ws_uri(self) -> str | None:
        return None

    @property
    def ipc_path(self) -> Path | None:
        return None

    def connect(self):
        self.__dict__.pop("tester", None)
        self._web3 = Web3(self.tester)
        # Handle disabling auto-mine if the user requested via config.
        if self.settings.auto_mine is False:
            self.auto_mine = False  # type: ignore[misc]

    def disconnect(self):
        # NOTE: This type ignore seems like a bug in pydantic.
        self._web3 = None
        self._evm_backend = None
        self.provider_settings = {}

        # Invalidate snapshots.
        self.chain_manager._snapshots[self.chain_id] = []

    def update_settings(self, new_settings: dict):
        self.provider_settings = {**self.provider_settings, **new_settings}
        self.disconnect()
        self.connect()

    def estimate_gas_cost(
        self, txn: "TransactionAPI", block_id: "BlockID | None" = None, **kwargs
    ) -> int:
        if isinstance(self.network.gas_limit, int):
            return self.network.gas_limit

        estimate_gas = self.web3.eth.estimate_gas

        # NOTE: Using JSON mode since used as request data.
        txn_dict = txn.model_dump(by_alias=True, mode="json", exclude={"gas_limit", "chain_id"})
        txn_data = cast(TxParams, txn_dict)

        try:
            return estimate_gas(txn_data, block_identifier=block_id)
        except (ValidationError, TransactionFailed, Web3ContractLogicError) as err:
            ape_err = self.get_virtual_machine_error(err, txn=txn, set_ape_traceback=False)
            gas_match = self._INVALID_NONCE_PATTERN.match(str(ape_err))
            if gas_match:
                # Sometimes, EthTester is confused about the sender nonce
                # during gas estimation. Retry using the "expected" gas
                # and then set it back.
                expected_nonce, actual_nonce = gas_match.groups()
                txn.nonce = int(expected_nonce)

                # NOTE: Using JSON mode since used as request data.
                txn_params: TxParams = cast(TxParams, txn.model_dump(by_alias=True, mode="json"))

                value = estimate_gas(txn_params, block_identifier=block_id)
                txn.nonce = int(actual_nonce)
                return value

            elif isinstance(ape_err, ContractLogicError):
                raise ape_err.with_ape_traceback() from err
            else:
                message = gas_estimation_error_message(ape_err)
                raise TransactionError(
                    message,
                    base_err=ape_err,
                    txn=txn,
                    source_traceback=lambda: ape_err.source_traceback,
                    set_ape_traceback=False,
                ) from ape_err

    @property
    def settings(self) -> EthTesterProviderConfig:
        return EthTesterProviderConfig.model_validate(
            {**self.config.provider.model_dump(), **self.provider_settings}
        )

    @property
    def supports_tracing(self) -> bool:
        return False

    @cached_property
    def chain_id(self) -> int:
        try:
            result = self.make_request("eth_chainId")
        except ProviderNotConnectedError:
            result = self.settings.chain_id

        return result

    @property
    def gas_price(self) -> int:
        return self.base_fee  # no miner tip

    @property
    def priority_fee(self) -> int:
        """Returns 0 because test chains do not care about priority fees."""
        return 0

    @property
    def base_fee(self) -> int:
        """
        EthTester does not implement RPC `eth_feeHistory`.
        Returns the last base fee on chain.
        """
        return self._get_last_base_fee()

    def send_call(
        self,
        txn: "TransactionAPI",
        block_id: "BlockID | None" = None,
        state: dict | None = None,
        **kwargs,
    ) -> HexBytes:
        # NOTE: Using JSON mode since used as request data.
        data = txn.model_dump(mode="json", exclude_none=True)

        state = kwargs.pop("state_override", None)
        call_kwargs: dict = {"block_identifier": block_id, "state_override": state}
        raise_on_revert = kwargs.get("raise_on_revert", txn.raise_on_revert)

        # Remove unneeded properties
        data.pop("gas", None)
        data.pop("gasLimit", None)
        data.pop("maxFeePerGas", None)
        data.pop("maxPriorityFeePerGas", None)
        data.pop("signature", None)

        tx_params = cast(TxParams, data)
        vm_err = None
        try:
            result = self.web3.eth.call(tx_params, **call_kwargs)
        except ValidationError as err:
            vm_err = VirtualMachineError(base_err=err)
            if raise_on_revert:
                raise vm_err from err

            elif isinstance(vm_err, ContractLogicError):
                # Convert revert reason back to bytes and use that as the return.
                return CallResult.from_revert(vm_err)

            else:
                result = HexBytes("0x")

        except (TransactionFailed, Web3ContractLogicError) as err:
            vm_err = self.get_virtual_machine_error(err, txn=txn, set_ape_traceback=False)
            if raise_on_revert:
                raise vm_err from err

            elif isinstance(vm_err, ContractLogicError):
                # Convert revert reason back to bytes and use that as the return.
                return CallResult.from_revert(vm_err)

            else:
                result = HexBytes("0x")

        self._increment_call_func_coverage_hit_count(txn)
        if vm_err:
            logger.error(vm_err)

        return HexBytes(result)

    def send_transaction(self, txn: "TransactionAPI") -> "ReceiptAPI":
        vm_err = None
        txn_dict = None
        try:
            txn_hash = self.tester.ethereum_tester.send_raw_transaction(
                to_hex(txn.serialize_transaction())
            )
        except (ValidationError, TransactionFailed, Web3ContractLogicError) as err:
            vm_err = self.get_virtual_machine_error(err, txn=txn, set_ape_traceback=False)
            if txn.raise_on_revert:
                raise vm_err from err
            else:
                txn_hash = to_hex(txn.txn_hash)

        required_confirmations = txn.required_confirmations or 0
        txn_dict = txn_dict or txn.model_dump(mode="json")

        # Signature is typically excluded from the model fields,
        # so we have to include it manually.
        txn_dict["signature"] = txn.signature

        if vm_err or not self.auto_mine:
            receipt_data = {
                **txn_dict,
                "block_number": -1,  # Not yet confirmed,
                "error": vm_err,
                "provider": self,
                "required_confirmations": required_confirmations,
                "status": (
                    TransactionStatusEnum.FAILING if vm_err else TransactionStatusEnum.NO_ERROR
                ),
                "txn_hash": txn_hash,
            }
            receipt = self.network.ecosystem.decode_receipt(receipt_data)
        else:
            receipt = self.get_receipt(
                txn_hash, required_confirmations=required_confirmations, transaction=txn_dict
            )

        # NOTE: Caching must happen before error enrichment.
        self.chain_manager.history.append(receipt)

        if receipt.failed:
            assert receipt.transaction.sender  # NOTE: mypy happy
            txn_dict["nonce"] = self.get_nonce(receipt.transaction.sender)
            txn_params = cast(TxParams, txn_dict)
            txn_dict.pop("signature", None)

            # Replay txn to get revert reason
            try:
                self.web3.eth.call(txn_params)
            except (ValidationError, TransactionFailed, Web3ContractLogicError) as err:
                vm_err = self.get_virtual_machine_error(err, txn=receipt, set_ape_traceback=False)
                receipt.error = vm_err
                if txn.raise_on_revert:
                    raise vm_err from err

            if txn.raise_on_revert:
                # If we get here, for some reason the tx-replay did not produce
                # a VM error.
                receipt.raise_for_status()

        if receipt.error:
            logger.error(receipt.error)

        return receipt

    def snapshot(self) -> "SnapshotID":
        return self.evm_backend.take_snapshot()

    def restore(self, snapshot_id: "SnapshotID"):
        # NOTE: Snapshot ID can be 0!
        if snapshot_id is None:
            return

        current_hash = self._get_latest_block_rpc().get("hash")
        if current_hash == snapshot_id:
            return

        try:
            return self.evm_backend.revert_to_snapshot(snapshot_id)
        except (HeaderNotFound, ValidationError):
            raise UnknownSnapshotError(snapshot_id)

    def set_timestamp(self, new_timestamp: int):
        current_timestamp = self.evm_backend.get_block_by_number("pending")["timestamp"]
        if new_timestamp == current_timestamp:
            # no change, return immediately
            return

        try:
            self.evm_backend.time_travel(new_timestamp)
        except ValidationError as err:
            pattern = re.compile(
                r"timestamp must be strictly later than parent, "
                r"but is 0 seconds before\.\n- child\s{2}: (\d*)\n- parent : (\d*)\.\s*"
            )
            if match := re.match(pattern, str(err)):
                if groups := match.groups():
                    if groups[0].strip() == groups[1].strip():
                        # Handle race condition when block headers are the same.
                        # Treat as noop, same as pre-check.
                        return

            raise ProviderError(f"Failed to time travel: {err}") from err

    def mine(self, num_blocks: int = 1):
        self.evm_backend.mine_blocks(num_blocks)

    def get_balance(self, address: AddressType, block_id: "BlockID | None" = None) -> int:
        # perf: Using evm_backend directly instead of going through web3.
        return self.evm_backend.get_balance(
            HexBytes(address), block_number="latest" if block_id is None else block_id
        )

    def get_nonce(self, address: AddressType, block_id: "BlockID | None" = None) -> int:
        return self.evm_backend.get_nonce(
            HexBytes(address), block_number="latest" if block_id is None else block_id
        )

    def get_contract_logs(self, log_filter: "LogFilter") -> Iterator["ContractLog"]:
        from_block = max(0, log_filter.start_block)

        if log_filter.stop_block is None:
            to_block = None
        else:
            latest_block = self._get_latest_block_rpc().get("number")
            to_block = (
                min(latest_block, log_filter.stop_block)
                if latest_block is not None
                else log_filter.stop_block
            )

        log_gen = self.tester.ethereum_tester.get_logs(
            address=log_filter.addresses,
            from_block=from_block,
            to_block=to_block,
            topics=log_filter.topic_filter,
        )
        yield from self.network.ecosystem.decode_logs(log_gen, *log_filter.events)

    def get_test_account(self, index: int) -> "TestAccountAPI":
        # NOTE: No need to cache here because it happens at the TestAccountManager already.
        try:
            private_key = self.evm_backend.account_keys[index]
        except IndexError as err:
            raise IndexError(f"No account at index '{index}'") from err

        address = private_key.public_key.to_canonical_address()
        return self.account_manager.init_test_account(
            index,
            cast(AddressType, to_hex(address)),
            str(private_key),
        )

    def add_account(self, private_key: str):
        self.evm_backend.add_account(private_key)

    def _get_last_base_fee(self) -> int:
        base_fee = self.evm_backend.get_block_by_number("pending").get("base_fee_per_gas", None)
        if base_fee is not None:
            return base_fee

        raise APINotImplementedError("No base fee found in block.")

    def get_transaction_trace(self, transaction_hash: str, **kwargs) -> "TraceAPI":
        if "call_trace_approach" not in kwargs:
            kwargs["call_trace_approach"] = TraceApproach.BASIC

        return EthTesterTransactionTrace(transaction_hash=transaction_hash, **kwargs)

    def get_virtual_machine_error(self, exception: Exception, **kwargs) -> VirtualMachineError:
        if isinstance(exception, ValidationError):
            match = self._CANNOT_AFFORD_GAS_PATTERN.match(str(exception))
            if match:
                txn_gas, bal = match.groups()
                sender = kwargs["txn"].sender
                new_message = (
                    f"Sender '{sender}' cannot afford txn gas {txn_gas} with account balance {bal}."
                )
                return VirtualMachineError(new_message, base_err=exception, **kwargs)

            else:
                return VirtualMachineError(base_err=exception, **kwargs)

        elif isinstance(exception, Web3ContractLogicError):
            # If the ape-solidity plugin is installed, we are able to enrich the data.
            message = exception.message
            raw_data = exception.data if isinstance(exception.data, str) else "0x"
            error = ContractLogicError(base_err=exception, revert_message=raw_data, **kwargs)
            enriched_error = self.compiler_manager.enrich_error(error)
            if is_0x_prefixed(enriched_error.message):
                # Failed to enrich. Use nicer message from web3.py.
                enriched_error.message = message or enriched_error.message

            return enriched_error

        elif isinstance(exception, TransactionFailed):
            err_message = str(exception).split("execution reverted: ")[-1] or None

            if err_message and err_message.startswith("b'") and err_message.endswith("'"):
                # Convert stringified bytes str like `"b'\\x82\\xb4)\\x00'"` to `"0x82b42900"`.
                # (Notice the `b'` is within the `"` on the first str).
                err_message = to_hex(literal_eval(err_message))

            err_message = TransactionError.DEFAULT_MESSAGE if err_message == "0x" else err_message
            contract_err = ContractLogicError(
                base_err=exception, revert_message=err_message, **kwargs
            )
            return self.compiler_manager.enrich_error(contract_err)

        else:
            return VirtualMachineError(base_err=exception, **kwargs)

    def _get_latest_block(self) -> BlockAPI:
        # perf: By-pass as much as possible since this is a common action.
        data = self._get_latest_block_rpc()
        return self.network.ecosystem.decode_block(data)

    def _get_latest_block_rpc(self) -> dict:
        return self.evm_backend.get_block_by_number("latest")


class EthTesterTransactionTrace(TransactionTrace):
    @cached_property
    def return_value(self) -> Any:
        # perf: skip trying anything else, because eth-tester doesn't
        # yet implement any tracing RPCs.
        init_kwargs = self._get_tx_calltree_kwargs()
        receipt = self.chain_manager.get_receipt(self.transaction_hash)
        init_kwargs["gas_cost"] = receipt.gas_used

        if not (abi := self.root_method_abi):
            return (None,)

        num_return = len(self.root_method_abi.outputs)

        # Figure out the 'returndata' using 'eth_call' RPC.
        tx = receipt.transaction.model_copy(update={"nonce": None})
        try:
            returndata = self.provider.send_call(tx, block_id=receipt.block_number)
        except ContractLogicError:
            # Unable to get the return value because even as a call, it fails.
            return tuple([None for _ in range(num_return)])

        return self._ecosystem.decode_returndata(abi, returndata)



from typing import Any

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import as_declarative


@as_declarative()
class Base:
    """
    Base class to generate ``__tablename__`` automatically
    """

    id: Any
    __name__: str

    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()
